//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam.h
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

#if !defined(LIBITZAM_ITZAM_H)
#define LIBITZAM_ITZAM_H

#if defined(__cplusplus)
extern "C" {
#endif

#if defined(_MSC_VER)
#pragma warning (disable : 4786 4244 4267 4101)
#endif

// Standard C library
#include <stdio.h>
#include <string.h>

#if defined(_MSC_VER)
typedef int bool;
static const int true  = 1;
static const int false = 0;

typedef unsigned __int64 uint64_t;
typedef __int64          int64_t;
typedef unsigned long    uint32_t;
typedef long             int32_t;
typedef unsigned short   uint16_t;
typedef short            int16_t;
typedef unsigned char    uint8_t;
typedef signed char      int8_t;
#else
#include <stdbool.h>
#include <stdint.h>
#endif

#if __WORDSIZE == 32
#define ITZAM32
#endif

//------------------------------------------------------------
// enumerated types
typedef enum
{
    ITZAM_ERROR_SIGNATURE,
    ITZAM_ERROR_VERSION,
    ITZAM_ERROR_64BIT_DB,
    ITZAM_ERROR_WRITE_FAILED,
    ITZAM_ERROR_OPEN_FAILED,
    ITZAM_ERROR_READ_FAILED,
    ITZAM_ERROR_CLOSE_FAILED,
    ITZAM_ERROR_SEEK_FAILED,
    ITZAM_ERROR_TELL_FAILED,
    ITZAM_ERROR_DUPE_REMOVE,
    ITZAM_ERROR_FLUSH_FAILED,
    ITZAM_ERROR_TOO_SMALL,
    ITZAM_ERROR_NULL_ALLOC,
    ITZAM_ERROR_PAGE_NOT_FOUND,
    ITZAM_ERROR_LOST_KEY,
    ITZAM_ERROR_KEY_NOT_WRITTEN,
    ITZAM_ERROR_KEY_SEEK_FAILED,
    ITZAM_ERROR_KEY_REMOVE_FAILED,
    ITZAM_ERROR_REC_SEEK_FAILED,
    ITZAM_ERROR_REC_REMOVE_FAILED,
    ITZAM_ERROR_DELLIST_NOT_READ,
    ITZAM_ERROR_DELLIST_NOT_WRITTEN,
    ITZAM_ERROR_ITERATOR_COUNT,
    ITZAM_ERROR_REWRITE_DELETED
} itzam_error;

typedef enum
{
    ITZAM_OKAY,
    ITZAM_AT_END,
    ITZAM_AT_BEGIN,
    ITZAM_UNKNOWN,
    ITZAM_NOT_FOUND,
    ITZAM_DUPLICATE,
    ITZAM_32BIT_OVERFLOW,
    ITZAM_DATA_WRITE_FAILED,
    ITZAM_REF_SIZE_MISMATCH
} itzam_state;

//------------------------------------------------------------
// simple types
#if defined(ITZAM32)
typedef int32_t itzam_ref;
typedef int32_t itzam_int;
#else
typedef int64_t itzam_ref;
typedef int64_t itzam_int;
#endif

//------------------------------------------------------------
// constants
static const itzam_ref ITZAM_NULL_REF     = -1;
static const itzam_ref ITZAM_NULL_KEY     = -1;

// flags for records
static const int32_t ITZAM_RECORD_IN_USE  =  0x00000001;
static const int32_t ITZAM_RECORD_DELLIST =  0x00000002;

//------------------------------------------------------------
// functions of this type must return:
//      < 0    key1 is before key2
//        0    keys are equal
//      > 0    key1 is after key2
typedef int itzam_key_comparator(const void * key1, const void * key2);

// function to select keys to be included in a subset
//      true    include key
//      false   don't include this key
typedef bool itzam_key_selector(const void * key);

// function type for error handler
typedef void itzam_error_handler(const char * function_name, itzam_error error);

// structures
//      this pragma is understood by MSVC and GCC
#pragma pack(1)

//------------------------------------------------------------
// variable-length data file structures

// record header, prefix for every record in a datafile
typedef struct
{
    uint32_t  m_flags;     // either ITZAM_RECORD_EMPTY or ITZAM_RECORD_IN_USE
    itzam_int m_length;    // record length
    itzam_int m_rec_size;  // number of bytes in use by actual data
}
itzam_record_header;

// data file header, prefix for entire data file
static const uint32_t ITZAM_DATAFILE_SIGNATURE = 0x4D5A5449; // ITZM

#if defined(ITZAM32)
static const uint32_t ITZAM_DATAFILE_VERSION   = 0x20030000; // 32.MM.MM.MM
#else
static const uint32_t ITZAM_DATAFILE_VERSION   = 0x40030000; // 64.MM.MM.MM
#endif

static const itzam_int ITZAM_DEL_LIST_BLOCK_SIZE    = 256;

typedef struct
{
    itzam_int   m_table_size; // number of entries in the table
}
itzam_dellist_header;

typedef struct
{
    itzam_ref   m_where;    // reference for this deleted record
    itzam_int   m_length;   // length of the deleted record
}
itzam_dellist_entry;

typedef struct
{
    uint32_t  m_signature;    // file type signature
    uint32_t  m_version;      // version of this file type
    itzam_ref m_dellist_ref;  // position of deleted list in file
}
itzam_datafile_header;

// working storage for a loaded datafile
typedef struct
{
    FILE *                m_file;           // file associated with this datafile
    itzam_datafile_header m_header;         // header information
    itzam_error_handler * m_error_handler;  // function to handle errors that occur
    itzam_dellist_header  m_dellist_header; // header for current deleted record list
    itzam_dellist_entry * m_dellist;        // pointer to list of deleted records
}
itzam_datafile;

//------------------------------------------------------------
// btree data structures

// forward decalaration
struct itzam_btree_page_t;
struct itzam_btree_page_entry_t;

// defines an entry in the page cache
typedef struct itzam_btree_page_entry_t
{
    struct itzam_btree_page_entry_t * m_prev; // previous entry in list
    struct itzam_btree_page_entry_t * m_next; // next entry in list
    struct itzam_btree_page_t       * m_page; // pointer to page in cache
}
itzam_btree_page_entry;

// page header, prefix for pages
typedef struct
{
    itzam_ref  m_where;     // location in page file
    itzam_ref  m_parent;    // parent in page file
    uint16_t   m_key_count; // actual  # of keys
}
itzam_btree_page_header;

// a page in the b-tree
typedef struct itzam_btree_page_t
{
    // the actual data for this reference
    uint8_t * m_data;

    // header information
    itzam_btree_page_header * m_header;

    // cache information
    itzam_btree_page_entry  * m_cache_entry;

    // these elements are pointers into m_data
    itzam_ref * m_keys;       // array of key values
    itzam_ref * m_refs;       // reference pointers
    itzam_ref * m_links;      // links to other pages
}
itzam_btree_page;

// btree file header, defines characteristics of a b-tree
static const uint32_t ITZAM_BTREE_VERSION = 0x00020000;

typedef struct
{
    uint32_t   m_version;         // version of this file structure
    uint32_t   m_sizeof_page;     // sizeof(itzam_btree_page) used in creating this btreefile
    uint16_t   m_order;           // number of keys per page
    itzam_int  m_count;           // counts number of active records
    itzam_int  m_ticker;          // counts the total number of new records added
    itzam_ref  m_where;           // pointer to location of this header
    itzam_ref  m_root_where;      // pointer to location of root reference in file
    itzam_ref  m_max_ref;         // the largest file reference used so far
}
itzam_btree_header;

// working storage for a loaded b-tree
typedef struct
{
    itzam_datafile *         m_datafile;       // file associated with this btree file
    itzam_btree_header       m_header;         // header data describing the tree structure
    bool                     m_free_datafile;  // free the datafile object when closed
    uint16_t                 m_links_size;     // number of links per page; order + 1; calculated at creation time
    uint16_t                 m_min_keys;       // minimum # of keys; order / 2; calculated at creation time
    itzam_key_comparator *   m_key_comparator; // function to compare keys

    // cache information
    uint16_t                 m_cache_max;      // maximum size of the cache
    uint16_t                 m_cache_size;     // number of entries in the cache
    itzam_btree_page_entry * m_cache_head;     // head (first) node in cache
    itzam_btree_page_entry * m_cache_tail;     // tail (last) node in cache
}
itzam_btree;

//------------------------------------------------------------
// btree iterator structures
typedef struct itzam_btree_iterator_entry_t
{
    // linked list entries
    struct itzam_btree_iterator_entry_t * m_prev;
    struct itzam_btree_iterator_entry_t * m_next;

    // data and key pointers
    itzam_ref m_key_ref;
    itzam_ref m_data_ref;
}
itzam_btree_iterator_entry;

typedef struct itzam_btree_iterator_t
{
    // btree associated with this iterator
    itzam_btree * m_btree;

    // ends of the list
    itzam_btree_iterator_entry * m_first;
    itzam_btree_iterator_entry * m_last;

    // current position in the list
    itzam_btree_iterator_entry * m_pointer;

    // number of entries in the list
    itzam_int m_count;
}
itzam_btree_iterator;

#pragma pack()

//----------------------------------------------------------
// retrieve a string that describes an itzam state
const char * itzam_get_state_description(itzam_state state);

// set the default error handler
void itzam_set_default_error_handler(itzam_error_handler * error_handler);

// get the default error handler
itzam_error_handler * itzam_get_default_error_handler();

// prototypes for default key comparisons
int itzam_default_key_comparator(const void * key1, const void * key2);

//----------------------------------------------------------
// prototypes for variable-length reference data file
itzam_state itzam_datafile_create(itzam_datafile * datafile,
                                  const char * filename);

itzam_state itzam_datafile_open(itzam_datafile * datafile,
                                const char * filename,
                                bool read_only);

itzam_state itzam_datafile_close(itzam_datafile * datafile);

int itzam_datafile_get_refbits(itzam_datafile * datafile);

float itzam_datafile_get_version(itzam_datafile * datafile);

void itzam_datafile_set_error_handler(itzam_datafile * datafile,
                                      itzam_error_handler * error_handler);

itzam_ref itzam_datafile_tell(itzam_datafile * datafile);

itzam_state itzam_datafile_seek(itzam_datafile * datafile,
                                itzam_ref pos);

itzam_state itzam_datafile_rewind(itzam_datafile * datafile);

itzam_state itzam_datafile_skip(itzam_datafile * datafile);

itzam_ref itzam_datafile_get_next_open(itzam_datafile * datafile,
                                       itzam_int length);

itzam_ref itzam_datafile_write(itzam_datafile * datafile,
                               const void * data,
                               itzam_int length,
                               itzam_ref where);

itzam_state itzam_datafile_rewrite(itzam_datafile * datafile,
                                   const void * data,
                                   itzam_int length,
                                   itzam_ref where);

itzam_state itzam_datafile_read(itzam_datafile * datafile,
                                void * data,
                                itzam_int max_length);

itzam_state itzam_datafile_read_alloc(itzam_datafile * datafile,
                                      void ** data,
                                      itzam_int * length);

itzam_state itzam_datafile_remove(itzam_datafile * datafile);

itzam_state itzam_datafile_commit(itzam_datafile * datafile);

//----------------------------------------------------------
// prototypes for btree file
itzam_state itzam_btree_create(itzam_btree * btree,
                               const char * filename,
                               itzam_key_comparator * key_comparator);

itzam_state itzam_btree_createdb(itzam_btree * btree,
                                 itzam_datafile * datafile,
                                 itzam_key_comparator * key_comparator);

itzam_state itzam_btree_open(itzam_btree * btree,
                             const char * filename,
                             itzam_key_comparator * key_comparator,
                             bool read_only);

itzam_state itzam_btree_opendb(itzam_btree * btree,
                               itzam_datafile * datafile,
                               itzam_key_comparator * key_comparator,
                               itzam_ref header_pos);

itzam_state itzam_btree_close(itzam_btree * btree);

void itzam_btree_set_error_handler(itzam_btree * btree,
                                   itzam_error_handler * error_handler);

itzam_state itzam_btree_insert(itzam_btree * btree,
                               const void * key,
                               itzam_int key_len,
                               itzam_ref reference);

itzam_state itzam_btree_insert_rec(itzam_btree * btree,
                                   const void * key,
                                   itzam_int key_len,
                                   const void * record,
                                   itzam_ref rec_len);

itzam_state itzam_btree_read_rec(itzam_btree * btree,
                                 const void * key,
                                 void * record,
                                 itzam_ref max_rec_len);

itzam_ref itzam_btree_find(itzam_btree * btree,
                           const void * key);

itzam_state itzam_btree_remove(itzam_btree * btree,
                               const void * key,
                               itzam_ref * ref);

//------------------------------------------------------------
// btree iterator methods
void itzam_btree_iterator_create(itzam_btree * btree,
                                 itzam_btree_iterator * iterator,
                                 itzam_key_selector * selector);

bool itzam_btree_iterator_is_valid(itzam_btree_iterator * iterator);

itzam_int itzam_btree_iterator_count(itzam_btree_iterator * iterator);

itzam_state itzam_btree_iterator_read_rec(itzam_btree_iterator * iterator,
                                          void * record,
                                          itzam_ref max_rec_len);

itzam_ref itzam_btree_iterator_get_rec_ref(itzam_btree_iterator * iterator);

itzam_ref itzam_btree_iterator_get_key_ref(itzam_btree_iterator * iterator);

itzam_state itzam_btree_iterator_move_prev(itzam_btree_iterator * iterator);

itzam_state itzam_btree_iterator_move_next(itzam_btree_iterator * iterator);

void itzam_btree_iterator_move_first(itzam_btree_iterator * iterator);

void itzam_btree_iterator_move_last(itzam_btree_iterator * iterator);

void itzam_btree_iterator_free(itzam_btree_iterator * iterator);

#if defined(__cplusplus)
}
#endif

#endif
