//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam_btree.c
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

// library header file
#include "itzam.h"

// Standard C library
#include <stdlib.h>

static const uint16_t PAGE_ORDER      =   24;
static const size_t   PAGE_CACHE_SIZE = 1024;

static itzam_state update_header(itzam_btree * btree)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // rewrite file header
    result = itzam_datafile_rewrite(btree->m_datafile,
                                    &btree->m_header,
                                    sizeof(btree->m_header),
                                    btree->m_header.m_where);

    // done
    return result;
}

static void set_page_pointers(itzam_btree_page * page, const itzam_btree * btree)
{
    // set pointers to appropriate bytes
    page->m_header = (itzam_btree_page_header *)page->m_data;

    page->m_keys   = (itzam_ref *)(page->m_data + sizeof(itzam_btree_page_header));

    page->m_refs   = (itzam_ref *)(page->m_data
                                      + sizeof(itzam_btree_page_header)
                                      + sizeof(itzam_ref) * btree->m_header.m_order);

    page->m_links  = (itzam_ref *)(page->m_data
                                      + sizeof(itzam_btree_page_header)
                                      + sizeof(itzam_ref) * btree->m_header.m_order
                                      + sizeof(itzam_ref) * btree->m_header.m_order);
}

static itzam_btree_page * alloc_page(const itzam_btree * btree)
{
    // return value
    int n;
    itzam_btree_page * page = NULL;

    // allocate page
    page = (itzam_btree_page *)malloc(sizeof(itzam_btree_page));

    if (page != NULL)
    {
        // allocate raw data
        page->m_data = (uint8_t *)malloc(btree->m_header.m_sizeof_page);

        if (page->m_data != NULL)
        {
            set_page_pointers(page,btree);

            // fill header with default values
            page->m_header->m_where     = ITZAM_NULL_REF;
            page->m_header->m_parent    = ITZAM_NULL_REF;
            page->m_header->m_key_count = 0;

            // initialize
            for (n = 0; n < btree->m_header.m_order; ++n)
            {
                page->m_keys[n]  = ITZAM_NULL_KEY;
                page->m_refs[n]  = ITZAM_NULL_REF;
                page->m_links[n] = ITZAM_NULL_REF;
            }

            page->m_links[btree->m_header.m_order] = ITZAM_NULL_REF;

            // initialize cache information (this page is not in cache yet, so null)
            page->m_cache_entry = NULL;
        }
        else
        {
            // error; clean up and return NULL
            free(page);
            page = NULL;
        }
    }

    // done
    return page;
}

static void free_page(itzam_btree_page * page)
{
    // only free if page is no longer in cache
    if ((page != NULL) && (page->m_cache_entry == NULL))
    {
        // clean up
        free(page->m_data);
        free(page);
    }
}

static void cache_delete(itzam_btree * btree, itzam_ref where)
{
    // search cache for this page; begin at the head
    itzam_btree_page_entry * entry = btree->m_cache_head;

    while (entry != NULL)
    {
        // is this the page we want?
        if (entry->m_page->m_header->m_where == where)
        {
            // is this the head?
            if (entry->m_prev == NULL)
                btree->m_cache_head = entry->m_next;
            else
                entry->m_prev->m_next = entry->m_next;

            // is this the tail?
            if (entry->m_next == NULL)
                btree->m_cache_tail = entry->m_prev;
            else
                entry->m_next->m_prev = entry->m_prev;

            // set page for deletion and delete it
            entry->m_page->m_cache_entry = NULL;
            free_page(entry->m_page);
            free(entry);

            // reduce cache size
            --btree->m_cache_size;

            break;
        }

        entry = entry->m_next;
    }
}

static bool cache_page(itzam_btree * btree, itzam_btree_page * page)
{
    // result
    bool result = false;

    // add page to cache
    if (btree->m_cache_head == NULL)
    {
        // this is the only entry in the list
        btree->m_cache_head = (itzam_btree_page_entry *)malloc(sizeof(itzam_btree_page_entry));

        if (btree->m_cache_head != NULL)
        {
            // set head as lone entry in cache
            btree->m_cache_head->m_prev = NULL;
            btree->m_cache_head->m_next = NULL;
            btree->m_cache_head->m_page = page;

            // tail and head are the same
            btree->m_cache_tail = btree->m_cache_head;

            // set count to 1
            btree->m_cache_size = 1;

            // it worked
            result = true;
        }
    }
    else
    {
        // add this entry to head of the list
        itzam_btree_page_entry * new_entry = (itzam_btree_page_entry *)malloc(sizeof(itzam_btree_page_entry));

        if (new_entry != NULL)
        {
            // set entry values
            new_entry->m_prev = NULL;
            new_entry->m_next = btree->m_cache_head;
            new_entry->m_page = page;

            // set head to point back new entry
            btree->m_cache_head->m_prev = new_entry;

            // new entry is now head
            btree->m_cache_head = new_entry;

            // have we exceeded to maximum size of the cache?
            if (btree->m_cache_size >= btree->m_cache_max)
            {
                itzam_btree_page_entry * dead_tail;

				// unlink tail from list
                btree->m_cache_tail->m_prev->m_next = NULL;

                // save dead tail pointer
                dead_tail = btree->m_cache_tail;

                // change tail
                btree->m_cache_tail = dead_tail->m_prev;

                // mark tail for deletion and delete
                dead_tail->m_page->m_cache_entry = NULL;
                free_page(dead_tail->m_page);

                // delete dead tail itself
                free(dead_tail);
            }
            else
            {
                // increment count
                ++btree->m_cache_size;
            }

            // it worked
            result = true;
        }
    }

    // set page pointer to its entry
    page->m_cache_entry = btree->m_cache_head;

    // done
    return result;
}

static itzam_btree_page * read_page_direct(itzam_btree * btree, itzam_ref where)
{
    // what we return
    itzam_btree_page * page = NULL;

    // position datafile
    if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,where))
    {
        // allocate a buffer
        page = alloc_page(btree);

        if (page != NULL)
        {
            // read the ref
            if (ITZAM_OKAY != itzam_datafile_read(btree->m_datafile,page->m_data,btree->m_header.m_sizeof_page))
            {
                // something broke; free pointer
                page->m_cache_entry = NULL;
                free_page(page);
                page = NULL;
            }
        }
    }

    // done
    return page;
}

// cen't be static because it's used by debug routines
itzam_btree_page * read_page(itzam_btree * btree, itzam_ref where)
{
    // what we return
    itzam_btree_page * page = NULL;
    
    // search cache for this page; begin at the head
    itzam_btree_page_entry * entry = btree->m_cache_head;

    while (entry != NULL)
    {
        // is this the page we want?
        if (entry->m_page->m_header->m_where == where)
        {
            // set the return
            page = entry->m_page;
            break;
        }

        entry = entry->m_next;
    }

    // if we didn't find the page
    if (page == NULL)
        page = read_page_direct(btree,where);

    // done
    return page;
}

static void update_max_ref(itzam_btree * btree, itzam_ref ref)
{
    // make sure we haven't exceeded system capacity
    if ((sizeof(long) == 4) && (ref > 0x7fffffff))
        btree->m_datafile->m_error_handler("update_max_ref",ITZAM_32BIT_OVERFLOW);

    // update the maximum ref for this tree, if necessary
    if (ref > btree->m_header.m_max_ref)
    {
        btree->m_header.m_max_ref = ref;
        update_header(btree);
    }
}

static itzam_state write_page(itzam_btree * btree, itzam_btree_page * page)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // does this page have a location, i.e., is it new?
    if (page->m_header->m_where == ITZAM_NULL_REF)
    {
        itzam_ref where;

        // assign root a file position
        page->m_header->m_where = itzam_datafile_get_next_open(btree->m_datafile,
                                                               btree->m_header.m_sizeof_page);

        // write root page
        where = itzam_datafile_write(btree->m_datafile,
                                     page->m_data,
                                     btree->m_header.m_sizeof_page,
                                     page->m_header->m_where);

        // check and update max reference
        update_max_ref(btree,where + btree->m_header.m_sizeof_page);

        // make sure
        if (where == page->m_header->m_where)
        {
            // assume we're writing the root if the header doesn't point to one
            if (btree->m_header.m_root_where == ITZAM_NULL_REF)
            {
                // update header with root data
                btree->m_header.m_root_where = where;
                update_header(btree);
            }

            // add page to cache
            if (cache_page(btree,page))
                result = ITZAM_OKAY;
        }
    }
    else
    {
        // rewrite page ref
        result = itzam_datafile_rewrite(btree->m_datafile,
                                        page->m_data,
                                        btree->m_header.m_sizeof_page,
                                        page->m_header->m_where);
    }

    // done
    return result;
}

static void set_btree_values(itzam_btree * btree,
                             itzam_key_comparator * key_comparator)
{
    // set array sizes
    btree->m_links_size = btree->m_header.m_order + 1;
    btree->m_min_keys   = btree->m_header.m_order / 2;

    // fill in initial values for page cache
    btree->m_cache_head = NULL;
    btree->m_cache_tail = NULL;
    btree->m_cache_size =  0;

    // if this number isn't big enough, pages can be pushed off
    //     the bottom of the cache during recursive operations!
    btree->m_cache_max  = PAGE_CACHE_SIZE;

    // set the error handler to the default
    btree->m_key_comparator = key_comparator;
}

itzam_state itzam_btree_create(itzam_btree * btree,
                               const char * filename,
                               itzam_key_comparator * key_comparator)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((btree != NULL) && (filename != NULL) && (key_comparator != NULL))
    {
        // create data file
        btree->m_datafile = malloc(sizeof(itzam_datafile));

        if (btree->m_datafile != NULL)
        {
            // remember to deallocate this datafile at closing
            btree->m_free_datafile = true;

            // create data file
            if (ITZAM_OKAY == itzam_datafile_create(btree->m_datafile,filename))
            {
                // versioning
                btree->m_header.m_version = ITZAM_BTREE_VERSION;

                // save order; calculate values derived from order
                btree->m_header.m_order = PAGE_ORDER;
                set_btree_values(btree,key_comparator);

                // set counts and tickers
                btree->m_header.m_count  = 0;
                btree->m_header.m_ticker = 0;

                // set explicit position for header
                btree->m_header.m_where = itzam_datafile_get_next_open(btree->m_datafile,sizeof(itzam_btree_header));

                // use a temporary root position (since we don't know it yet)
                btree->m_header.m_root_where = 0;

                // these values ensure that a database is handled properly on different platforms
                btree->m_header.m_sizeof_page = sizeof(itzam_btree_page_header)
                                              + sizeof(itzam_ref) * btree->m_header.m_order
                                              + sizeof(itzam_ref) * btree->m_header.m_order
                                              + sizeof(itzam_ref) * btree->m_links_size;

                // write header for first time (lacks root pointer info, but needs to occupy space in the file)
                if (btree->m_header.m_where == itzam_datafile_write(btree->m_datafile,
                                                                    &btree->m_header,
                                                                    sizeof(btree->m_header),
                                                                    btree->m_header.m_where))
                {
                    // create root page
                    itzam_btree_page * root = alloc_page(btree);

                    if (root != NULL)
                    {
                        // assign root a file position
                        root->m_header->m_where = itzam_datafile_get_next_open(btree->m_datafile,
                                                                               btree->m_header.m_sizeof_page);

                        if (root->m_header->m_where != ITZAM_NULL_REF)
                        {
                            // write root page
                            btree->m_header.m_root_where = itzam_datafile_write(btree->m_datafile,
                                                                                root->m_data,
                                                                                btree->m_header.m_sizeof_page,
                                                                                root->m_header->m_where);

                            // make certain that the root was written where we expect it to be
                            if ((btree->m_header.m_root_where == root->m_header->m_where)
                            &&  (btree->m_header.m_root_where != ITZAM_NULL_REF))
                            {
                                // set maximum ref to root location
                                btree->m_header.m_max_ref = btree->m_header.m_root_where + btree->m_header.m_sizeof_page;

                                // rewrite file header
                                result = update_header(btree);
                            }
                        }
                    }
                }
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_createdb(itzam_btree * btree,
                                 itzam_datafile * datafile,
                                 itzam_key_comparator * key_comparator)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((btree != NULL) && (datafile != NULL) && (key_comparator != NULL))
    {
        // create data file
        btree->m_header.m_where = itzam_datafile_get_next_open(btree->m_datafile,sizeof(itzam_btree_header));

        if (ITZAM_NULL_REF != btree->m_header.m_where)
        {
            // remember datafile
            btree->m_datafile = datafile;
            btree->m_free_datafile = false; // we didn't allocate it, so don't free it

            // versioning
            btree->m_header.m_version = ITZAM_BTREE_VERSION;

            // save order; calculate values derived from order
            btree->m_header.m_order = PAGE_ORDER;
            set_btree_values(btree,key_comparator);

            // set counts and tickers
            btree->m_header.m_count  = 0;
            btree->m_header.m_ticker = 0;

            // use a temporary root position (since we don't know it yet)
            btree->m_header.m_root_where = 0;

            // these values ensure that a database is handled
            btree->m_header.m_sizeof_page = sizeof(itzam_btree_page_header)
                                          + sizeof(itzam_ref) * btree->m_header.m_order
                                          + sizeof(itzam_ref) * btree->m_header.m_order
                                          + sizeof(itzam_ref) * btree->m_links_size;

            // write header for first time (lacks root pointer info, but needs to occupy space in the file)
            if (btree->m_header.m_where == itzam_datafile_write(btree->m_datafile,
                                                                &btree->m_header,
                                                                sizeof(btree->m_header),
                                                                btree->m_header.m_root_where))
            {
                // create root page
                itzam_btree_page * root = alloc_page(btree);

                if (root != NULL)
                {
                    // assign root a file position
                    root->m_header->m_where = itzam_datafile_get_next_open(btree->m_datafile,
                                                                           btree->m_header.m_sizeof_page);

                    if (root->m_header->m_where != ITZAM_NULL_REF)
                    {
                        // write root page
                        btree->m_header.m_root_where = itzam_datafile_write(btree->m_datafile,
                                                                            root->m_data,
                                                                            btree->m_header.m_sizeof_page,
                                                                            root->m_header->m_where);

                        // make certain that the root was written where we expect it to be
                        if ((btree->m_header.m_root_where == root->m_header->m_where)
                        &&  (btree->m_header.m_root_where != ITZAM_NULL_REF))
                        {
                            // set maximum ref to roo location
                            btree->m_header.m_max_ref = btree->m_header.m_root_where + btree->m_header.m_sizeof_page;

                            // rewrite file header
                            result = update_header(btree);
                        }
                    }
                }
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_open(itzam_btree * btree,
                             const char * filename,
                             itzam_key_comparator * key_comparator,
                             bool read_only)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // create data file
    btree->m_datafile = malloc(sizeof(itzam_datafile));

    if ((btree->m_datafile != NULL) && (key_comparator != NULL))
    {
        // make sure the arguments make sense
        if ((btree != NULL) && (filename != NULL))
        {
            // open data file
            if (ITZAM_OKAY == itzam_datafile_open(btree->m_datafile, filename, read_only))
            {
                // if given a filename, assumes first record is header
                if (ITZAM_OKAY == itzam_datafile_rewind(btree->m_datafile))
                {
                    if (ITZAM_OKAY == itzam_datafile_read(btree->m_datafile,&btree->m_header,sizeof(btree->m_header)))
                    {
                        // verify version
                        if (btree->m_header.m_version == ITZAM_BTREE_VERSION)
                        {
                            set_btree_values(btree,key_comparator);
                            btree->m_free_datafile = true;
                            result = ITZAM_OKAY;
                        }
                        else
                            result = ITZAM_UNKNOWN;
                    }
                }
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_opendb(itzam_btree * btree,
                               itzam_datafile * datafile,
                               itzam_key_comparator * key_comparator,
                               itzam_ref header_pos)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((btree != NULL) && (datafile != NULL) && (key_comparator != NULL))
    {
        // remember datafile
        btree->m_datafile = datafile;

        // open data file
        if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile, header_pos))
        {
            // if given a filename, assumes first record is header
            if (ITZAM_OKAY == itzam_datafile_read(btree->m_datafile,&btree->m_header,sizeof(btree->m_header)))
            {
                // verify version
                if (btree->m_header.m_version == ITZAM_BTREE_VERSION)
                {
                    set_btree_values(btree,key_comparator);
                    btree->m_free_datafile = false;

                    if ((sizeof(long) == 4) && (btree->m_header.m_max_ref > 0x7fffffff))
                        result = ITZAM_UNKNOWN;
                    else
                        result = ITZAM_OKAY;
                }
                else
                    result = ITZAM_UNKNOWN;
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_close(itzam_btree * btree)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // free the page cache
    itzam_btree_page_entry * entry = btree->m_cache_head;

    while (entry != NULL)
    {
        // save pointer to next entry
        itzam_btree_page_entry * next = entry->m_next;

        // set page for deletion and remove it
        entry->m_page->m_cache_entry = NULL;
        free_page(entry->m_page);

        // now free this cache entry
        free(entry);

        // next iteration
        entry = next;
    }

    // make sure the arguments make sense
    if (btree != NULL)
    {
        itzam_datafile_close(btree->m_datafile);

        if (btree->m_free_datafile)
            free(btree->m_datafile);

        result = ITZAM_OKAY;
    }

    // done
    return result;
}

// assign an error handler
void itzam_btree_set_error_handler(itzam_btree * btree,
                                   itzam_error_handler * error_handler)
{
    itzam_datafile_set_error_handler(btree->m_datafile,error_handler);
}

// compare a given key to one read from the file
static int compare_keys(itzam_btree * btree, const void * key, itzam_ref key2_ref)
{
    // read key record
    int comp = 0;
    void * key_data;
    itzam_ref length;

    if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,key2_ref))
    {
        if (ITZAM_OKAY == itzam_datafile_read_alloc(btree->m_datafile,&key_data,&length))
        {
            // compare
            comp = btree->m_key_comparator(key,key_data);
            free(key_data);
        }
        else
            btree->m_datafile->m_error_handler("compare_keys",ITZAM_ERROR_LOST_KEY);
    }
    else
        btree->m_datafile->m_error_handler("compare_keys",ITZAM_ERROR_KEY_SEEK_FAILED);

    return comp;
}

// structure used to return search information
typedef struct
{
    itzam_btree_page * m_page;
    int                m_index;
    bool               m_found;
} search_result;

// search for leaf node in which we insert the given key
static search_result search(itzam_btree * btree, itzam_btree_page * page, const void * key)
{
    // what we return
    search_result result;

    int index = 0;

    // if page is empty, we didn't find the key
    if (page->m_header->m_key_count == 0)
    {
        result.m_page  = page;
        result.m_index = index;
        result.m_found = false;
    }
    else
    {
        // loop through keys
        while (index < page->m_header->m_key_count)
        {
            int comp = compare_keys(btree,key,page->m_keys[index]);

            // do we move on, or have we found it?
            if (comp > 0)
                ++index;
            else
            {
                if (comp == 0)
                {
                    result.m_page  = page;
                    result.m_index = index;
                    result.m_found = true;
                    return result;
                }

                break;
            }
        }

        // if we're in a leaf, the key hasn't been found
        if (page->m_links[index] == ITZAM_NULL_REF)
        {
            result.m_page  = page;
            result.m_index = index;
            result.m_found = false;
        }
        else
        {
            // read and search next page
            itzam_btree_page * next_page = read_page(btree,page->m_links[index]);

            // search next page
            result = search(btree,next_page,key);
        }
    }

    // done
    return result;
}

// promote key by creating new root
static void promote_root(itzam_btree * btree,
                         itzam_ref key,
                         itzam_ref ref,
                         itzam_btree_page * page_before,
                         itzam_btree_page * page_after)
{
    // create a new root page
    itzam_btree_page * new_root = alloc_page(btree);

    // add key and links to root
    new_root->m_keys[0]  = key;
    new_root->m_refs[0]  = ref;
    new_root->m_links[0] = page_before->m_header->m_where;
    new_root->m_links[1] = page_after->m_header->m_where;
    new_root->m_header->m_key_count = 1;

    // update header with new root
    btree->m_header.m_root_where = ITZAM_NULL_REF;

    // write new root to tree
    write_page(btree,new_root);

    // update children
    page_before->m_header->m_parent = new_root->m_header->m_where;
    page_after->m_header->m_parent  = new_root->m_header->m_where;

    write_page(btree,page_before);
    write_page(btree,page_after);
}

// promote key into parent
static void promote_internal(itzam_btree * btree,
                             itzam_btree_page * page_insert,
                             itzam_ref key_ref,
                             itzam_ref ref,
                             itzam_ref link)
{
    if (page_insert->m_header->m_key_count == btree->m_header.m_order)
    {
        int nt  = 0;
        int ni  = 0;
        int insert_index = 0;

        // read key data
        void * key_data;
        itzam_ref length;
        itzam_btree_page * page_sibling;
        itzam_btree_page * child;

        // temporary array
        itzam_ref * temp_keys  = (itzam_ref *)malloc(sizeof(itzam_ref) * (btree->m_header.m_order + 1));
        itzam_ref * temp_refs  = (itzam_ref *)malloc(sizeof(itzam_ref) * (btree->m_header.m_order + 1));
        itzam_ref * temp_links = (itzam_ref *)malloc(sizeof(itzam_ref) * (btree->m_header.m_order + 2));

        temp_links[0] = page_insert->m_links[0];

        // read key data
        if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,key_ref))
        {
            if (ITZAM_OKAY != itzam_datafile_read_alloc(btree->m_datafile,&key_data,&length))
                btree->m_datafile->m_error_handler("promote_internal",ITZAM_ERROR_LOST_KEY);
        }
        else
            btree->m_datafile->m_error_handler("promote_internal",ITZAM_ERROR_KEY_SEEK_FAILED);

        // find insertion point
        while ((insert_index < page_insert->m_header->m_key_count)
                 && (compare_keys(btree,key_data,page_insert->m_keys[insert_index]) >= 0))
            ++insert_index;

        // free key data
        free(key_data);

        // store new info
        temp_keys[insert_index]      = key_ref;
        temp_refs[insert_index]      = ref;
        temp_links[insert_index + 1] = link;

        // copy existing keys
        while (ni < btree->m_header.m_order)
        {
            if (ni == insert_index)
                ++nt;

            temp_keys[nt]      = page_insert->m_keys[ni];
            temp_refs[nt]      = page_insert->m_refs[ni];
            temp_links[nt + 1] = page_insert->m_links[ni + 1];

            ++ni;
            ++nt;
        }

        // generate a new leaf node
        page_sibling = alloc_page(btree);
        page_sibling->m_header->m_parent = page_insert->m_header->m_parent;

        // clear key counts
        page_insert->m_header->m_key_count = 0;
        page_sibling->m_header->m_key_count = 0;
        page_insert->m_links[0] = temp_links[0];

        // copy keys from temp to pages
        for (ni = 0; ni < btree->m_min_keys; ++ni)
        {
            page_insert->m_keys[ni]      = temp_keys[ni];
            page_insert->m_refs[ni]      = temp_refs[ni];
            page_insert->m_links[ni + 1] = temp_links[ni + 1];

            ++page_insert->m_header->m_key_count;
        }

        page_sibling->m_links[0] = temp_links[btree->m_min_keys + 1];

        for (ni = btree->m_min_keys + 1; ni <= btree->m_header.m_order; ++ni)
        {
            page_sibling->m_keys[ni - 1 - btree->m_min_keys] = temp_keys[ni];
            page_sibling->m_refs[ni - 1 - btree->m_min_keys] = temp_refs[ni];
            page_sibling->m_links[ni - btree->m_min_keys]    = temp_links[ni + 1];

            ++page_sibling->m_header->m_key_count;
        }

        // replace unused entries with null
        for (ni = btree->m_min_keys; ni < btree->m_header.m_order; ++ni)
        {
            page_insert->m_keys[ni]      = ITZAM_NULL_KEY;
            page_insert->m_refs[ni]      = ITZAM_NULL_REF;
            page_insert->m_links[ni + 1] = ITZAM_NULL_REF;
        }

        // write pages
        write_page(btree,page_insert);
        write_page(btree,page_sibling);

        // update parent links in child nodes
        for (ni = 0; ni <= page_sibling->m_header->m_key_count; ++ni)
        {
            child = read_page(btree,page_sibling->m_links[ni]);

            if (child != NULL)
            {
                child->m_header->m_parent = page_sibling->m_header->m_where;
                write_page(btree,child);
            }
            else
                btree->m_datafile->m_error_handler("promote_internal",ITZAM_ERROR_PAGE_NOT_FOUND);
        }

        // promote key and pointer
        if (page_insert->m_header->m_parent == ITZAM_NULL_REF)
        {
            // create a new root
            promote_root(btree,
                         temp_keys[btree->m_min_keys],
                         temp_refs[btree->m_min_keys],
                         page_insert,
                         page_sibling);
        }
        else
        {
            // read parent and promote key
            itzam_btree_page * parent_page = read_page(btree,page_insert->m_header->m_parent);

            promote_internal(btree,
                             parent_page,
                             temp_keys[btree->m_min_keys],
                             temp_refs[btree->m_min_keys],
                             page_sibling->m_header->m_where);
        }

        // release resources
        free(temp_keys);
        free(temp_refs);
        free(temp_links);
    }
    else
    {
        int n, insert_index = 0;

        // read key data
        void * key_data;
        itzam_ref length;

        if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,key_ref))
        {
            if (ITZAM_OKAY != itzam_datafile_read_alloc(btree->m_datafile,&key_data,&length))
                btree->m_datafile->m_error_handler("promote_internal",ITZAM_ERROR_LOST_KEY);
        }
        else
            btree->m_datafile->m_error_handler("promote_internal",ITZAM_ERROR_KEY_SEEK_FAILED);

        // find insertion point
        while ((insert_index < page_insert->m_header->m_key_count)
                 && (compare_keys(btree,key_data,page_insert->m_keys[insert_index]) >= 0))
            ++insert_index;

        // free key data
        free(key_data);

        // shift keys right
        for (n = page_insert->m_header->m_key_count; n > insert_index; --n)
        {
            page_insert->m_keys[n]      = page_insert->m_keys[n - 1];
            page_insert->m_refs[n]      = page_insert->m_refs[n - 1];
            page_insert->m_links[n + 1] = page_insert->m_links[n];
        }

        page_insert->m_keys[insert_index]      = key_ref;
        page_insert->m_refs[insert_index]      = ref;
        page_insert->m_links[insert_index + 1] = link;

        ++page_insert->m_header->m_key_count;

        // write updated page
        write_page(btree,page_insert);
    }
}

// internal function to write keys
static void write_key(itzam_btree * btree,
                      search_result * insert_info,
                      const void * key,
                      itzam_ref key_len,
                      itzam_ref ref)
{
    itzam_btree_page * page_sibling;
    itzam_btree_page * page_parent;

    // write the key to the file, if necessary
    itzam_ref key_ref = itzam_datafile_write(btree->m_datafile,key,key_len,ITZAM_NULL_REF);

    // check and update max reference
    update_max_ref(btree,key_ref + key_len);

    if (key_ref >= 0)
    {
        // check to see if page is full
        if (insert_info->m_page->m_header->m_key_count == btree->m_header.m_order)
        {
            int nt, ni;

            // temporary arrays
            itzam_ref * temp_keys = (itzam_ref *)malloc(sizeof(itzam_ref) * (btree->m_header.m_order + 1));
            itzam_ref * temp_refs = (itzam_ref *)malloc(sizeof(itzam_ref) * (btree->m_header.m_order + 1));

            // store new item
            temp_keys[insert_info->m_index] = key_ref;
            temp_refs[insert_info->m_index] = ref;

            // copy entries from insertion page to temps
            // MAYBE: change these to use pointers instead of indexes?
            nt = 0; // index for temp arrays
            ni = 0; // index for insertion page

            while (ni < btree->m_header.m_order)
            {
                // skip over inserted data
                if (ni == insert_info->m_index)
                    ++nt;

                // copy data
                temp_keys[nt] = insert_info->m_page->m_keys[ni];
                temp_refs[nt] = insert_info->m_page->m_refs[ni];

                // next one
                ++ni;
                ++nt;
            }

            // create a new leaf
            page_sibling = alloc_page(btree);
            page_sibling->m_header->m_parent = insert_info->m_page->m_header->m_parent;

            // clear key counts
            insert_info->m_page->m_header->m_key_count = 0;
            page_sibling->m_header->m_key_count        = 0;

            // copy keys from temp to pages
            for (ni = 0; ni < btree->m_min_keys; ++ni)
            {
                insert_info->m_page->m_keys[ni] = temp_keys[ni];
                insert_info->m_page->m_refs[ni] = temp_refs[ni];

                ++insert_info->m_page->m_header->m_key_count;
            }

            for (ni = btree->m_min_keys + 1; ni <= btree->m_header.m_order; ++ni)
            {
                page_sibling->m_keys[ni - 1 - btree->m_min_keys] = temp_keys[ni];
                page_sibling->m_refs[ni - 1 - btree->m_min_keys] = temp_refs[ni];

                ++page_sibling->m_header->m_key_count;
            }

            // replace remaining entries with null
            for (ni = btree->m_min_keys; ni < btree->m_header.m_order; ++ni)
            {
                insert_info->m_page->m_keys[ni] = ITZAM_NULL_KEY;
                insert_info->m_page->m_refs[ni] = ITZAM_NULL_REF;
            }

            // write pages
            write_page(btree,insert_info->m_page);
            write_page(btree,page_sibling);

            // promote key and its pointer
            if (insert_info->m_page->m_header->m_parent == ITZAM_NULL_REF)
            {
                // creating a new root page
                promote_root(btree,
                             temp_keys[btree->m_min_keys],
                             temp_refs[btree->m_min_keys],
                             insert_info->m_page,
                             page_sibling);
            }
            else
            {
                // read parent
                page_parent = read_page(btree,insert_info->m_page->m_header->m_parent);

                // promote key into parent page
                promote_internal(btree,
                                 page_parent,
                                 temp_keys[btree->m_min_keys],
                                 temp_refs[btree->m_min_keys],
                                 page_sibling->m_header->m_where);
            }

            // release sibling page
            free(temp_keys);
            free(temp_refs);
        }
        else
        {
            int n;

            // move keys to make room for new one
            for (n = insert_info->m_page->m_header->m_key_count; n > insert_info->m_index; --n)
            {
                insert_info->m_page->m_keys[n] = insert_info->m_page->m_keys[n - 1];
                insert_info->m_page->m_refs[n] = insert_info->m_page->m_refs[n - 1];
            }

            insert_info->m_page->m_keys[insert_info->m_index] = key_ref;
            insert_info->m_page->m_refs[insert_info->m_index] = ref;

            ++insert_info->m_page->m_header->m_key_count;

            // write updated page
            write_page(btree,insert_info->m_page);
        }
    }
    else
        btree->m_datafile->m_error_handler("write_key",ITZAM_ERROR_KEY_NOT_WRITTEN);
}

itzam_state itzam_btree_insert(itzam_btree * btree,
                                 const void * key,
                                 itzam_ref key_len,
                                 itzam_ref ref)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((btree != NULL) && (key != NULL))
    {
        // search for the key
        itzam_btree_page * root = read_page(btree,btree->m_header.m_root_where);

        if (root != NULL)
        {
            search_result insert_info = search(btree,root,key);

            if (!insert_info.m_found) // || btree->m_header.m_allows_dupes)
            {
                write_key(btree,&insert_info,key,key_len,ref);
                ++btree->m_header.m_count;
                ++btree->m_header.m_ticker;

                result = update_header(btree);
            }
            else
            {
                result = ITZAM_DUPLICATE;
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_insert_rec(itzam_btree * btree,
                                   const void * key,
                                   itzam_int key_len,
                                   const void * record,
                                   itzam_int rec_len)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;
    itzam_ref ref = ITZAM_NULL_REF;

    // make sure the arguments make sense
    if ((btree != NULL) && (key != NULL))
    {
        // search for the key
        itzam_btree_page * root = read_page(btree,btree->m_header.m_root_where);

        if (root != NULL)
        {
            search_result insert_info = search(btree,root,key);

            if (!insert_info.m_found) // || btree->m_header.m_allows_dupes)
            {
                ref = itzam_datafile_write(btree->m_datafile,record,rec_len,ITZAM_NULL_REF); 
                
                if (ref == ITZAM_NULL_REF)
                    result = ITZAM_DATA_WRITE_FAILED;
                else
                {
                    write_key(btree,&insert_info,key,key_len,ref);
                    ++btree->m_header.m_count;
                    ++btree->m_header.m_ticker;
                    result = update_header(btree);
                }
            }
            else
            {
                result = ITZAM_DUPLICATE;
            }
        }
    }

    // done
    return result;
}

itzam_state itzam_btree_read_rec(itzam_btree * btree,
                                 const void * key,
                                 void * record,
                                 itzam_ref max_rec_len)
{
    itzam_state result = ITZAM_UNKNOWN;

    itzam_ref ref = itzam_btree_find(btree,key);

    if (ref != ITZAM_NULL_REF)
    {
        result = itzam_datafile_seek(btree->m_datafile,ref);

        if (result == ITZAM_OKAY)
            result = itzam_datafile_read(btree->m_datafile,record,max_rec_len);
    }
    else
        result = ITZAM_NOT_FOUND;

    return result;
}

itzam_ref itzam_btree_find(itzam_btree * btree, const void * key)
{
    // what we return
    itzam_ref result = ITZAM_NULL_REF;

    // make sure the arguments make sense
    if (btree != NULL)
    {
        // search for the key
        itzam_btree_page * root = read_page(btree,btree->m_header.m_root_where);

        if (root != NULL)
        {
            search_result find_info = search(btree,root,key);

            if (find_info.m_found)
                result = find_info.m_page->m_refs[find_info.m_index];
        }
    }

    // done
    return result;
}

static void redistribute(itzam_btree * btree, int index, itzam_btree_page * page_before, itzam_btree_page * page_parent, itzam_btree_page * page_after)
{
    if ((btree != NULL) && (page_before != NULL) && (page_parent != NULL) && (page_after != NULL))
    {
        // check for leaf page
        if (page_before->m_links[0] == ITZAM_NULL_REF)
        {
            if (page_before->m_header->m_key_count > page_after->m_header->m_key_count)
            {
				int n;

                // move a key from page_before to page_after
                for (n = page_after->m_header->m_key_count; n > 0; --n)
                {
                    page_after->m_keys[n] = page_after->m_keys[n - 1];
                    page_after->m_refs[n] = page_after->m_refs[n - 1];
                }

                // store parent separator in page_after
                page_after->m_keys[0] = page_parent->m_keys[index];
                page_after->m_refs[0] = page_parent->m_refs[index];

                // increment page_after key count
                ++page_after->m_header->m_key_count;

                // decrement page_before key count
                --page_before->m_header->m_key_count;

                // move last key in page_before to page_parent as separator
                page_parent->m_keys[index] = page_before->m_keys[page_before->m_header->m_key_count];
                page_parent->m_refs[index] = page_before->m_refs[page_before->m_header->m_key_count];

                // clear last key in page_before
                page_before->m_keys[page_before->m_header->m_key_count] = ITZAM_NULL_KEY;
                page_before->m_refs[page_before->m_header->m_key_count] = ITZAM_NULL_REF;
            }
            else
            {
                int n;

                // add parent key to lesser page
                page_before->m_keys[page_before->m_header->m_key_count] = page_parent->m_keys[index];
                page_before->m_refs[page_before->m_header->m_key_count] = page_parent->m_refs[index];

                // increment page_before key count
                ++page_before->m_header->m_key_count;

                // move first key in page_after to page_parent as separator
                page_parent->m_keys[index] = page_after->m_keys[0];
                page_parent->m_refs[index] = page_after->m_refs[0];

                // decrement page_after key count
                --page_after->m_header->m_key_count;

                // move a key from page_after to page_before
                for (n = 0; n < page_after->m_header->m_key_count; ++n)
                {
                    page_after->m_keys[n] = page_after->m_keys[n + 1];
                    page_after->m_refs[n] = page_after->m_refs[n + 1];
                }

                // clear last key in page_after
                page_after->m_keys[n] = ITZAM_NULL_KEY;
                page_after->m_refs[n] = ITZAM_NULL_REF;
            }
        }
        else
        {
            itzam_btree_page * page_child;

            if (page_before->m_header->m_key_count > page_after->m_header->m_key_count)
            {
                int n;

                // move a key from page_before to page_after
                for (n = page_after->m_header->m_key_count; n > 0; --n)
                {
                    page_after->m_keys[n]      = page_after->m_keys[n - 1];
                    page_after->m_refs[n]      = page_after->m_refs[n - 1];
                    page_after->m_links[n + 1] = page_after->m_links[n];
                }

                page_after->m_links[1] = page_after->m_links[0];

                // store page_parent separator key in page_after
                page_after->m_keys[0]  = page_parent->m_keys[index];
                page_after->m_refs[0]  = page_parent->m_refs[index];
                page_after->m_links[0] = page_before->m_links[page_before->m_header->m_key_count];

                // update child link
                page_child = read_page(btree,page_after->m_links[0]);

                if (page_child != NULL)
                {
                    page_child->m_header->m_parent = page_after->m_header->m_where;
                    write_page(btree,page_child);
                }
                else
                    btree->m_datafile->m_error_handler("redistribute",ITZAM_ERROR_PAGE_NOT_FOUND);

                // increment page_after key count
                ++page_after->m_header->m_key_count;

                // decrement page_before key count
                --page_before->m_header->m_key_count;

                // move last key in page_before to page_parent as separator
                page_parent->m_keys[index] = page_before->m_keys[page_before->m_header->m_key_count];
                page_parent->m_refs[index] = page_before->m_refs[page_before->m_header->m_key_count];

                // clear last key in page_before
                page_before->m_keys[page_before->m_header->m_key_count]      = ITZAM_NULL_KEY;
                page_before->m_refs[page_before->m_header->m_key_count]      = ITZAM_NULL_REF;
                page_before->m_links[page_before->m_header->m_key_count + 1] = ITZAM_NULL_REF;
            }
            else
            {
                int n;

                // store page_parent separator key in page_before
                page_before->m_keys[page_before->m_header->m_key_count]      = page_parent->m_keys[index];
                page_before->m_refs[page_before->m_header->m_key_count]      = page_parent->m_refs[index];
                page_before->m_links[page_before->m_header->m_key_count + 1] = page_after->m_links[0];

                // update child link
                page_child = read_page(btree,page_after->m_links[0]);

                if (page_child != NULL)
                {
                    // page_child->m_header->m_parent = page_after->m_header->m_where;
                    page_child->m_header->m_parent = page_before->m_header->m_where;
                    write_page(btree,page_child);
                }
                else
                    btree->m_datafile->m_error_handler("redistribute",ITZAM_ERROR_PAGE_NOT_FOUND);

                // increment page_before key count
                ++page_before->m_header->m_key_count;

                // move last key in page_after to page_parent as separator
                page_parent->m_keys[index] = page_after->m_keys[0];
                page_parent->m_refs[index] = page_after->m_refs[0];

                // decrement page_after key count
                --page_after->m_header->m_key_count;

                // move a key from page_after to page_before
                for (n = 0; n < page_after->m_header->m_key_count; ++n)
                {
                    page_after->m_keys[n]  = page_after->m_keys[n + 1];
                    page_after->m_refs[n]  = page_after->m_refs[n + 1];
                    page_after->m_links[n] = page_after->m_links[n + 1];
                }

                page_after->m_links[n] = page_after->m_links[n + 1];

                // clear last key in page_after
                page_after->m_keys[n]      = ITZAM_NULL_KEY;
                page_after->m_refs[n]      = ITZAM_NULL_REF;
                page_after->m_links[n + 1] = ITZAM_NULL_REF;
            }
        }

        // write updated pages
        write_page(btree,page_before);
        write_page(btree,page_parent);
        write_page(btree,page_after);
    }
}

// prototype for concatenate
static void adjust_tree(itzam_btree * btree, itzam_btree_page * page);

static void concatenate(itzam_btree * btree, int index, itzam_btree_page * page_before, itzam_btree_page * page_parent, itzam_btree_page * page_after)
{
    int n, n2;

    // move separator key from page_parent into page_before
    page_before->m_keys[page_before->m_header->m_key_count]      = page_parent->m_keys[index];
    page_before->m_refs[page_before->m_header->m_key_count]      = page_parent->m_refs[index];
    page_before->m_links[page_before->m_header->m_key_count + 1] = page_after->m_links[0];

    // increment page_before key count
    ++page_before->m_header->m_key_count;

    // delete separator from page_parent
    --page_parent->m_header->m_key_count;

    for (n = index; n < page_parent->m_header->m_key_count; ++n)
    {
        page_parent->m_keys[n]      = page_parent->m_keys[n + 1];
        page_parent->m_refs[n]      = page_parent->m_refs[n + 1];
        page_parent->m_links[n + 1] = page_parent->m_links[n + 2];
    }

    // clear unused key from parent
    page_parent->m_keys[n]      = ITZAM_NULL_KEY;
    page_parent->m_refs[n]      = ITZAM_NULL_REF;
    page_parent->m_links[n + 1] = ITZAM_NULL_REF;

    // copy keys from page_after to page_before
    n2 = 0;
    n = page_before->m_header->m_key_count;

    // combine pages
    while (n2 < page_after->m_header->m_key_count)
    {
        ++page_before->m_header->m_key_count;

        page_before->m_keys[n]      = page_after->m_keys[n2];
        page_before->m_refs[n]      = page_after->m_refs[n2];
        page_before->m_links[n + 1] = page_after->m_links[n2 + 1];

        ++n2;
        ++n;
    }

    // delete page_after
    itzam_datafile_seek(btree->m_datafile,page_after->m_header->m_where);
    itzam_datafile_remove(btree->m_datafile);
    cache_delete(btree,page_after->m_header->m_where);

    // is this an inner page?
    if (page_before->m_links[0] != ITZAM_NULL_REF)
    {
        // adjust child pointers
		    itzam_btree_page * page_child;

		    for (n = 0; n <= page_before->m_header->m_key_count; ++n)
        {
            // read child
            page_child = read_page(btree,page_before->m_links[n]);

            // make sure the child was actually read
            if (page_child != NULL)
            {
                page_child->m_header->m_parent = page_before->m_header->m_where;
                write_page(btree,page_child);
            }
            else
                btree->m_datafile->m_error_handler("concatenate",ITZAM_ERROR_PAGE_NOT_FOUND);
        }
    }

    // write page_before and parent
    if (page_parent->m_header->m_key_count == 0)
    {
        // update before page with old parent's parent
        page_before->m_header->m_parent = page_parent->m_header->m_parent;
        write_page(btree,page_before);

        if (page_before->m_header->m_parent == ITZAM_NULL_REF)
        {
            // update the header if this is the new root
            btree->m_header.m_root_where = page_before->m_header->m_where;
            update_header(btree);
        }
        else
        {
            // read parents's parent
            itzam_btree_page * parents_parent = read_page(btree,page_parent->m_header->m_parent);

            // find parent reference and replace with before page
            if (parents_parent != NULL)
            {
                for (n = 0; n < btree->m_links_size; ++n)
                {
                    if (parents_parent->m_links[n] == page_parent->m_header->m_where)
                    {
                        parents_parent->m_links[n] = page_before->m_header->m_where;
                        write_page(btree,parents_parent);
                    }
                }
            }
        }

        // remove empty parent page
        itzam_datafile_seek(btree->m_datafile,page_parent->m_header->m_where);
        itzam_datafile_remove(btree->m_datafile);
        cache_delete(btree,page_parent->m_header->m_where);
    }
    else
    {
        // write pages
        write_page(btree,page_parent);
        write_page(btree,page_before);

        // reset root page if needed
        if (page_parent->m_header->m_parent == ITZAM_NULL_REF)
            btree->m_header.m_root_where = page_parent->m_header->m_where;

        // if parent is too small, adjust
        if (page_parent->m_header->m_key_count < btree->m_min_keys)
            adjust_tree(btree,page_parent);
    }
}

static void adjust_tree(itzam_btree * btree, itzam_btree_page * page)
{
    if ((btree != NULL) && (page != NULL) && (page->m_header->m_parent != ITZAM_NULL_REF))
    {
        // get parent page
        itzam_btree_page * page_parent = read_page(btree,page->m_header->m_parent);

        if (page_parent != NULL)
        {
            itzam_btree_page * page_sibling_after  = NULL;
            itzam_btree_page * page_sibling_before = NULL;

            // find pointer to page
            int n = 0;

            while (page_parent->m_links[n] != page->m_header->m_where)
                ++n;

            // read sibling pages
            if (n < page_parent->m_header->m_key_count)
                page_sibling_after = read_page(btree,page_parent->m_links[n+1]);

            if (n > 0)
                page_sibling_before = read_page(btree,page_parent->m_links[n-1]);

            // figure out what to do
            if (page_sibling_before != NULL)
            {
                --n;

                if (page_sibling_before->m_header->m_key_count > btree->m_min_keys)
                    redistribute(btree,n,page_sibling_before,page_parent,page);
                else
                    concatenate(btree,n,page_sibling_before,page_parent,page);
            }
            else
            {
                if (page_sibling_after != NULL)
                {
                    if (page_sibling_after->m_header->m_key_count > btree->m_min_keys)
                        redistribute(btree,n,page,page_parent,page_sibling_after);
                    else
                        concatenate(btree,n,page,page_parent,page_sibling_after);
                }
            }
        }
    }
}

itzam_state itzam_btree_remove(itzam_btree * btree, const void * key, itzam_ref * ref)
{
    // what we return
    itzam_state state, result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((btree != NULL) && (key != NULL))
    {
        // search for the key
        itzam_btree_page * root = read_page(btree,btree->m_header.m_root_where);

        if (root != NULL)
        {
            // search for key
            search_result remove_info = search(btree,root,key);

            if (remove_info.m_found)
            {
                // delete key record
                if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,remove_info.m_page->m_keys[remove_info.m_index]))
                {
                    if (ITZAM_OKAY != itzam_datafile_remove(btree->m_datafile))
                        btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_KEY_REMOVE_FAILED);
                }
                else
                    btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_KEY_SEEK_FAILED);
                    

                // remove data associated with key
                if (ref != NULL)
                    *ref = remove_info.m_page->m_refs[remove_info.m_index];
                else
                {
                    // if we don't send back the reference, delete the data record from the btree itself
                    if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,remove_info.m_page->m_refs[remove_info.m_index]))
                    {
                        if (ITZAM_OKAY != itzam_datafile_remove(btree->m_datafile))
                            btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_REC_REMOVE_FAILED);
                    }
                    else
                        btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_REC_SEEK_FAILED);
                }

                // is this a leaf node?
                if (remove_info.m_page->m_links[0] == ITZAM_NULL_REF)
                {
                    int n;

                    // removing key from leaf
                    --remove_info.m_page->m_header->m_key_count;

                    // slide keys left over removed one
                    for (n = remove_info.m_index; n < remove_info.m_page->m_header->m_key_count; ++n)
                    {
                        remove_info.m_page->m_keys[n] = remove_info.m_page->m_keys[n + 1];
                        remove_info.m_page->m_refs[n] = remove_info.m_page->m_refs[n + 1];
                    }

                    remove_info.m_page->m_keys[remove_info.m_page->m_header->m_key_count] = ITZAM_NULL_KEY;
                    remove_info.m_page->m_refs[remove_info.m_page->m_header->m_key_count] = ITZAM_NULL_REF;

                    // write changed record
                    write_page(btree,remove_info.m_page);

                    // adjust the tree, if needed
                    if (remove_info.m_page->m_header->m_key_count < btree->m_min_keys)
                        adjust_tree(btree,remove_info.m_page);

                    result = ITZAM_OKAY;
                }
                else // removing from an internal page
                {
                    // get the successor page
                    itzam_btree_page * page_successor = read_page(btree,remove_info.m_page->m_links[remove_info.m_index + 1]);

                    if (page_successor != NULL)
                    {
                        int n;

                        while (page_successor->m_links[0] != ITZAM_NULL_REF)
                        {
                            itzam_btree_page * next_successor = read_page(btree,page_successor->m_links[0]);

                            // check for null page in case of corrupted index
                            if (next_successor != NULL)
                            {
                                free_page(page_successor);
                                page_successor = next_successor;
                            }
                            else
                                btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_PAGE_NOT_FOUND);
                        }

                        // first key is the "swappee"
                        remove_info.m_page->m_keys[remove_info.m_index] = page_successor->m_keys[0];
                        remove_info.m_page->m_refs[remove_info.m_index] = page_successor->m_refs[0];

                        // remove swapped key from successor page
                        --page_successor->m_header->m_key_count;

                        for (n = 0; n < page_successor->m_header->m_key_count; ++n)
                        {
                            page_successor->m_keys[n]      = page_successor->m_keys[n + 1];
                            page_successor->m_refs[n]      = page_successor->m_refs[n + 1];
                            page_successor->m_links[n + 1] = page_successor->m_links[n + 2];
                        }

                        page_successor->m_keys[page_successor->m_header->m_key_count]      = ITZAM_NULL_KEY;
                        page_successor->m_refs[page_successor->m_header->m_key_count]      = ITZAM_NULL_REF;
                        page_successor->m_links[page_successor->m_header->m_key_count + 1] = ITZAM_NULL_REF;

                        // write modified records
                        write_page(btree,remove_info.m_page);
                        write_page(btree,page_successor);

                        // adjust tree for leaf node
                        if (page_successor->m_header->m_key_count < btree->m_min_keys)
                            adjust_tree(btree,page_successor);

                        result = ITZAM_OKAY;
                    }
                }

                // decrement number of records in file
                --btree->m_header.m_count;
                update_header(btree);
            }
            else
            {
                // btree->m_datafile->m_error_handler("itzam_btree_remove",ITZAM_ERROR_PAGE_NOT_FOUND);
                result = ITZAM_NOT_FOUND;
            }
        }
    }

    // done
    return result;
}

//------------------------------------------------------------
// btree iterator methods
static void itzam_btree_iterator_create_recurse(itzam_btree * btree,
                                                itzam_ref page_ref,
                                                itzam_btree_iterator * iterator,
                                                itzam_key_selector * selector)
{
    // Visual C++ can't handle an inline definition of this
    bool add_it;

    // read the page
    itzam_btree_page * page = read_page_direct(btree, page_ref);

    // recurse its keys
    if ((page != NULL) && (page->m_header->m_key_count > 0))
    {
        size_t n = 0;

        while (n < page->m_header->m_key_count)
        {
            itzam_btree_iterator_entry * new_entry;

            // recurse page with keys before this one (assuming this isn't a leaf page)
            if (page->m_links[n] != ITZAM_NULL_REF)
                itzam_btree_iterator_create_recurse(btree,page->m_links[n],iterator,selector);

            // should this item by included?
            add_it = (selector == NULL);

            // if we have a a selector, see if this key belongs in the set
            if (selector != NULL)
            {
                void * key;
                itzam_ref length;

                // get the key
                if (ITZAM_OKAY == itzam_datafile_seek(btree->m_datafile,page->m_keys[n]))
                {
                    if (ITZAM_OKAY == itzam_datafile_read_alloc(btree->m_datafile,&key,&length))
                    {
                        // decide whether or not to add it based on what the selector says
                        add_it = selector(key);

                        // clean up
                        free(key);
                    }
                    else
                        btree->m_datafile->m_error_handler("itzam_btree_iterator_create_recurse",ITZAM_ERROR_LOST_KEY);
                }
                else
                    btree->m_datafile->m_error_handler("itzam_btree_iterator_create_recurse",ITZAM_ERROR_KEY_SEEK_FAILED);

            }

            if (add_it)
            {
                // allocate a new entry in the list
                new_entry = (itzam_btree_iterator_entry *)malloc(sizeof(itzam_btree_iterator_entry));

                // if the list is empty, make this the first entry
                if (iterator->m_first == NULL)
                {
                    iterator->m_first   = new_entry; // iterator->m_pointer;
                    iterator->m_pointer = new_entry;
                    new_entry->m_prev   = NULL;
                }
                else
                {
                    // add to end of list (if the list isn't empty)
                    iterator->m_pointer->m_next = new_entry;
                    new_entry->m_prev = iterator->m_pointer;
                }

                // fill in the new list entry
                new_entry->m_next     = NULL;
                new_entry->m_key_ref  = page->m_keys[n];
                new_entry->m_data_ref = page->m_refs[n];

                // make the new entry the current pointer (end of list)
                iterator->m_pointer = new_entry;

                // this is also the last item in the list
                iterator->m_last = new_entry;

                // count the number of keys found
                ++iterator->m_count;
            }

            // next page
            ++n;
        }

        // recurse last page (with all keys after the ones on this page)
        if (page->m_links[page->m_header->m_key_count] != ITZAM_NULL_REF)
            itzam_btree_iterator_create_recurse(btree,page->m_links[page->m_header->m_key_count],iterator,selector);
    }

    // free the page
    free_page(page);
}


void itzam_btree_iterator_create(itzam_btree * btree,
                                 itzam_btree_iterator * iterator,
                                 itzam_key_selector * selector)
{
    if ((btree != NULL) && (iterator != NULL))
    {
        // keep track of the associated btree
        iterator->m_btree = btree;

        // we count the number of keys found
        iterator->m_count = 0;

        // the list is empty
        iterator->m_first   = NULL;
        iterator->m_pointer = NULL;

        // begin at the root
        itzam_btree_iterator_create_recurse(btree,btree->m_header.m_root_where,iterator,selector);

        // initialize current position to beginning of list
        iterator->m_pointer = iterator->m_first;

        // final integrity check
        if (selector == NULL)
        {
            if (iterator->m_count != btree->m_header.m_count)
                btree->m_datafile->m_error_handler("itzam_btree_iterator_create",
                                                   ITZAM_ERROR_ITERATOR_COUNT);
        }
    }
}

bool itzam_btree_iterator_is_valid(itzam_btree_iterator * iterator)
{
    return ((iterator != NULL) && (iterator->m_pointer != NULL));
}

itzam_int itzam_btree_iterator_count(itzam_btree_iterator * iterator)
{
    return iterator->m_count;
}

itzam_state itzam_btree_iterator_read_rec(itzam_btree_iterator * iterator,
                                          void * record,
                                          itzam_ref max_rec_len)
{
    itzam_datafile_seek(iterator->m_btree->m_datafile,iterator->m_pointer->m_data_ref);
    return itzam_datafile_read(iterator->m_btree->m_datafile,record,max_rec_len);
}

itzam_ref itzam_btree_iterator_get_rec_ref(itzam_btree_iterator * iterator)
{
    itzam_ref result = ITZAM_NULL_REF;

    if ((iterator != NULL) && (iterator->m_pointer != NULL))
        result = iterator->m_pointer->m_data_ref;

    return result;
}

itzam_ref itzam_btree_iterator_get_key_ref(itzam_btree_iterator * iterator)
{
    itzam_ref result = ITZAM_NULL_REF;

    if ((iterator != NULL) && (iterator->m_pointer != NULL))
        result = iterator->m_pointer->m_key_ref;

    return result;
}

itzam_state itzam_btree_iterator_move_prev(itzam_btree_iterator * iterator)
{
    itzam_state result = ITZAM_UNKNOWN;

    // make certain we're working with a valid iterator
    if ((iterator != NULL) && (iterator->m_pointer != NULL))
    {
        if (iterator->m_pointer->m_prev == NULL)
            result = ITZAM_AT_BEGIN;
        else
        {
            iterator->m_pointer = iterator->m_pointer->m_prev;
            result = ITZAM_OKAY;
        }
    }

    return result;
}

itzam_state itzam_btree_iterator_move_next(itzam_btree_iterator * iterator)
{
    itzam_state result = ITZAM_UNKNOWN;

    // make certain we're working with a valid iterator
    if ((iterator != NULL) && (iterator->m_pointer != NULL))
    {
        if (iterator->m_pointer->m_next == NULL)
            result = ITZAM_AT_END;
        else
        {
            iterator->m_pointer = iterator->m_pointer->m_next;
            result = ITZAM_OKAY;
        }
    }

    return result;
}

void itzam_btree_iterator_move_first(itzam_btree_iterator * iterator)
{
    // make certain we're working with a valid iterator
    if ((iterator != NULL) && (iterator->m_pointer != NULL))
        iterator->m_pointer = iterator->m_first;
}

void itzam_btree_iterator_move_last(itzam_btree_iterator * iterator)
{
    // make certain we're working with a valid iterator
    if ((iterator != NULL) && (iterator->m_pointer != NULL))
        iterator->m_pointer = iterator->m_last;
}

void itzam_btree_iterator_free(itzam_btree_iterator * iterator)
{
    // make certain we're working with a valid iterator
    if ((iterator != NULL) && (iterator->m_pointer != NULL))
    {
        // loop through the list
        while (iterator->m_first != NULL)
        {
            // save pointer to next entry in list
            itzam_btree_iterator_entry * next = iterator->m_first->m_next;

            // remove the head of the list
            free(iterator->m_first);

            // now the next is the first
            iterator->m_first = next;
        }

        // doing this marks the iterator as deleted, and renders the other methods inoperative
        iterator->m_pointer = NULL;
    }
}
