//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam_btree_debug.c
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

// library header file
#include "itzam_debug.h"

// Standard C library
#include <stdlib.h>

itzam_btree_page * read_page(itzam_btree * btree, int64_t where);

void dump_page(itzam_btree * btree, itzam_btree_page * page)
{
    int n;

    fprintf(stderr,"-->> PAGE DUMP\n");
    fprintf(stderr,"m_data *      = %8p, m_header * = %8p\n",(void *)page->m_data,(void *)page->m_header);
    fprintf(stderr,"m_where       = %ld (0x%lx)\n",page->m_header->m_where,page->m_header->m_where);
    fprintf(stderr,"m_parent      = %ld (0x%lx)\n",page->m_header->m_parent,page->m_header->m_parent);
    fprintf(stderr,"m_key_count   = %u (0x%x)\n",page->m_header->m_key_count,page->m_header->m_key_count);

    for (n = 0; n < btree->m_header.m_order; ++n)
    {
        fprintf(stderr,"m_links[%04d] = %ld (0x%lx)\n",n,page->m_links[n],page->m_links[n]);
        fprintf(stderr,"m_keys [%04d] = %ld (0x%lx)\n",n,page->m_keys[n],page->m_keys[n]);
        fprintf(stderr,"m_refs [%04d] = %ld (0x%lx)\n",n,page->m_refs[n],page->m_refs[n]);
    }

    fprintf(stderr,"m_links[%04d] = %ld (0x%lx)\n",n,page->m_links[n],page->m_links[n]);
    fprintf(stderr,"<<--\n");
}

static void dump_btree_recursive(itzam_btree * btree, itzam_btree_page * page)
{
    int n;
    dump_page(btree,page);

    if (page->m_header->m_key_count > 0)
    {
        for (n = 0; n <= page->m_header->m_key_count; ++n)
        {
            if (page->m_links[n] != ITZAM_NULL_REF)
            {
                itzam_btree_page * temp = read_page(btree,page->m_links[n]);
                dump_btree_recursive(btree,temp);
                // free_page(temp);
            }
        }
    }
}

static void print_header(const itzam_btree * btree, FILE * output)
{
    fprintf(output,"btree header information\n------------------------------------------------------------\n");
    fprintf(output,"btree *                = 0x%8p\n",(void *)btree);
    fprintf(output,"m_header.m_sizeof_page = %d\n",btree->m_header.m_sizeof_page);
    fprintf(output,"m_header.m_order       = %d\n",btree->m_header.m_order);
    fprintf(output,"m_header.m_count       = %ld\n",btree->m_header.m_count);
    fprintf(output,"m_header.m_ticker      = %ld\n",btree->m_header.m_ticker);
    fprintf(output,"m_header.m_where       = %ld (0x%lx)\n",btree->m_header.m_where,btree->m_header.m_where);
    fprintf(output,"m_header.m_root_where  = %ld (0x%lx)\n",btree->m_header.m_root_where,btree->m_header.m_root_where);
    fprintf(output,"m_links_size           = %d\n",btree->m_links_size);
    fprintf(output,"m_min_keys             = %d\n",btree->m_min_keys);
    fprintf(output,"------------------------------------------------------------\n");
}

void dump_btree(itzam_btree * btree)
{
    itzam_btree_page * root;

    fprintf(stderr,"---> BEGIN BTREE ANALYSIS\n");
    print_header(btree,stderr);

    root = read_page(btree,btree->m_header.m_root_where);
    dump_btree_recursive(btree,root);

    fprintf(stderr,"<--- END BTREE DUMP\n");
}

struct anal_data
{
    FILE *      output;
    uint32_t    total_avail;
    uint32_t    total_inuse;
    double      balance;
    double      branch_count;
    uint32_t    height;
};

static void analyze_page(itzam_btree * btree, itzam_btree_page * page, struct anal_data * data)
{
    data->total_avail += btree->m_header.m_order;
    data->total_inuse += page->m_header->m_key_count;

    if (page->m_links[0] != ITZAM_NULL_REF)
    {
        data->balance += (double)page->m_header->m_key_count / (double)btree->m_header.m_order;
        ++data->branch_count;
    }

    fprintf(data->output,".");
}

static void analyze_btree_recursive(itzam_btree * btree, itzam_btree_page * page, struct anal_data * data)
{
    int n;

    analyze_page(btree,page,data);

    if (page->m_header->m_key_count > 0)
    {
        for (n = 0; n <= page->m_header->m_key_count; ++n)
        {
            if (page->m_links[n] != ITZAM_NULL_REF)
            {
                itzam_btree_page * temp = read_page(btree,page->m_links[n]);
                analyze_btree_recursive(btree,temp,data);
            }
        }
    }
}

static void analyze_btree_height(itzam_btree * btree, struct anal_data * data)
{
    itzam_btree_page * page = read_page(btree,btree->m_header.m_root_where);

    data->height = 1;

    while (page->m_links[0] != ITZAM_NULL_REF)
    {
        ++data->height;
        page = read_page(btree,page->m_links[0]);
    }
}

void analyze_btree(itzam_btree * btree, FILE * output)
{
    struct anal_data data;
    itzam_btree_page * root;
    itzam_record_header header;
    itzam_int del_count = 0;

    data.output       = output;
    data.total_avail  = 0;
    data.total_inuse  = 0;
    data.balance      = 0.0;
    data.branch_count = 1;
    data.height       = 0;

    fprintf(output,"---> BEGIN BTREE ANALYSIS\n\n");
    print_header(btree,output);

    fprintf(output,"\nanalyzing\n");
    root = read_page(btree,btree->m_header.m_root_where);
    analyze_btree_recursive(btree,root,&data);
    analyze_btree_height(btree,&data);
    fprintf(output,"\ncomplete\n");

    fprintf(output,"\noverall statistics:\n");
    fprintf(output,"----------------------------------------------------------------------\n");
    fprintf(output," key slots available = %8u\n",data.total_avail);
    fprintf(output,"    key slots in use = %8u\n",data.total_inuse);
    fprintf(output,"    total # of pages = %8u\n",(unsigned int)data.branch_count);
    fprintf(output,"         keys in use = %8.2f%%\n",100.0 * (double)data.total_inuse / (double)data.total_avail);
    fprintf(output,"overall tree balance = %8.2f%% (50%% or better is good in large trees)\n",100.0 * data.balance / (double)data.branch_count);
    fprintf(output,"         tree height = %8u\n",data.height);
    fprintf(output,"----------------------------------------------------------------------\n");
    
    // list datafile information
    analyze_datafile(btree->m_datafile,output);

    fprintf(output,"<--- END BTREE ANALYSIS\n\n");
}

void cache_dump(itzam_btree * btree)
{
    itzam_btree_page_entry * entry;
    fprintf(stderr,"cache dump\n");

    entry = btree->m_cache_head;

    while (entry != NULL)
    {
        fprintf(stderr,"  found entry %p, page %lu\n",(void *)entry,entry->m_page->m_header->m_where);
        entry = entry->m_next;
    }

    fprintf(stderr,"end cache dump\n");
}
