//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam_data.c
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

// library header file
#include "itzam.h"

#include <stdlib.h>

static itzam_state read_dellist(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    if (0 == fseek(datafile->m_file,datafile->m_header.m_dellist_ref + sizeof(itzam_record_header),SEEK_SET))
    {
        // read the header
        if (1 == fread(&datafile->m_dellist_header,sizeof(itzam_dellist_header),1,datafile->m_file))
        {
            itzam_int size = sizeof(itzam_dellist_entry) * datafile->m_dellist_header.m_table_size;
            datafile->m_dellist = (itzam_dellist_entry *)malloc(size);

            if (datafile->m_dellist != NULL)
            {
                if (1 == fread(datafile->m_dellist,size,1,datafile->m_file))
                    result = ITZAM_OKAY;
            }
        }
    }
    
    if (result != ITZAM_OKAY)
        datafile->m_error_handler("read_dellist",ITZAM_ERROR_DELLIST_NOT_READ);
    
    return result; 
}

static itzam_state write_dellist(itzam_datafile * datafile, bool has_grown)
{
    // what we return
    itzam_int   size   = sizeof(itzam_dellist_entry) * datafile->m_dellist_header.m_table_size;
    itzam_state result = ITZAM_UNKNOWN;
    itzam_record_header header;

    if (has_grown)
    {
        // remove old record if there was one
        if (datafile->m_header.m_dellist_ref != ITZAM_NULL_REF)
        {
            if (0 == fseek(datafile->m_file,datafile->m_header.m_dellist_ref,SEEK_SET))
            {
                // read the header
                if (1 == fread(&header,sizeof(header),1,datafile->m_file))
                {
                    // change record header; make this record the head of the deleted list
                    header.m_flags &= ~ITZAM_RECORD_IN_USE;
                    header.m_flags &= ~ITZAM_RECORD_DELLIST;

                    // move to beginning of record header again and rewrite it
                    fseek(datafile->m_file,datafile->m_header.m_dellist_ref,SEEK_SET);
                    fwrite(&header,sizeof(header),1,datafile->m_file);
                } 
            }
        }

        // explicitly append; we can't use write because it might try to change the
        //      deleted list while we're saving it
        if (0 == fseek(datafile->m_file,0,SEEK_END))
        {
            datafile->m_header.m_dellist_ref = ftell(datafile->m_file);

            if (datafile->m_header.m_dellist_ref > 0)
            {
                // write new record header
                itzam_record_header rec_head;
                
                rec_head.m_flags  = ITZAM_RECORD_IN_USE | ITZAM_RECORD_DELLIST;
                rec_head.m_length = sizeof(itzam_dellist_header) + size;
                rec_head.m_rec_size = rec_head.m_length;
                
                if (1 == fwrite(&rec_head,sizeof(itzam_record_header),1,datafile->m_file))
                {
                    // write the list header
                    if (1 == fwrite(&datafile->m_dellist_header,sizeof(itzam_dellist_header),1,datafile->m_file))
                    {
                        // write the list
                        if (1 == fwrite(datafile->m_dellist,size,1,datafile->m_file))
                        {
                            // update the header with the pointer to the new deleted record list
                            if (0 == fseek(datafile->m_file,0,SEEK_SET))
                            {
                                if (1 == fwrite(&datafile->m_header,sizeof(itzam_datafile_header),1,datafile->m_file))
                                    result = ITZAM_OKAY;
                            }
                        }
                    }
                }
            }
        }
    }
    else
    {
        if (0 == fseek(datafile->m_file,datafile->m_header.m_dellist_ref + sizeof(itzam_record_header),SEEK_SET))
        {
            // write the header
            if (1 == fwrite(&datafile->m_dellist_header,sizeof(itzam_dellist_header),1,datafile->m_file))
            {
                // write the list
                if (1 == fwrite(datafile->m_dellist,size,1,datafile->m_file))
                    result = ITZAM_OKAY;
            }
        }
    }

    if (result != ITZAM_OKAY)
        datafile->m_error_handler("read_dellist",ITZAM_ERROR_DELLIST_NOT_WRITTEN);

    return result;
}

// prototypes for variable-length record data file
itzam_state itzam_datafile_create(itzam_datafile * datafile, const char * filename)
{
#if _MSC_VER >= 1400
    errno_t err = 0;
#endif

    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // set default error handler
        datafile->m_error_handler = itzam_get_default_error_handler();

        // fill header
        datafile->m_header.m_signature     = ITZAM_DATAFILE_SIGNATURE;
        datafile->m_header.m_version       = ITZAM_DATAFILE_VERSION;
        datafile->m_header.m_dellist_ref   = ITZAM_NULL_REF;

        // assumes that datafile is a clean structure (i.e., it isn't open already)
#if _MSC_VER >= 1400
        err = fopen_s(&datafile->m_file,filename,"w+b"); // create or truncate; binary mode for non-Unix systems

        if (err != 0)
            datafile->m_file = NULL;
#else
        datafile->m_file = fopen(filename,"w+b"); // create or truncate; binary mode for non-Unix systems
#endif

        if (datafile->m_file != NULL)
        {
            // write header
            if (1 == fwrite(&datafile->m_header,sizeof(itzam_datafile_header),1,datafile->m_file))
                result = ITZAM_OKAY;
        }
    }

    // done
    return result;
}

itzam_state itzam_datafile_open(itzam_datafile * datafile, const char * filename, bool read_only)
{
#if _MSC_VER >= 1400
    errno_t err = 0;
#endif

    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // set default error handler
    datafile->m_error_handler  = itzam_get_default_error_handler();

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // assumes that datafile is a clean structure (i.e., it isn't open already)
#if _MSC_VER >= 1400
        if (read_only)
            err = fopen_s(&datafile->m_file,filename,"rb"); // create or truncate; binary mode for non-Unix systems
        else
            err = fopen_s(&datafile->m_file,filename,"r+b"); // create or truncate; binary mode for non-Unix systems

        if (err != 0)
            datafile->m_file = NULL;
#else
        if (read_only)
            datafile->m_file = fopen(filename,"rb"); // create or truncate; binary mode for non-Unix systems
        else
            datafile->m_file = fopen(filename,"r+b"); // create or truncate; binary mode for non-Unix systems
#endif

        if (datafile->m_file != NULL)
        {
            // read the deleted number
            if (1 == fread(&datafile->m_header,sizeof(itzam_datafile_header),1,datafile->m_file))
            {
                // verify signature and version
                if (datafile->m_header.m_signature == ITZAM_DATAFILE_SIGNATURE)
                {
                    if (datafile->m_header.m_version == ITZAM_DATAFILE_VERSION)
                    {
                        // read deleted list, if any
                        if (datafile->m_header.m_dellist_ref != ITZAM_NULL_REF)
                            result = read_dellist(datafile);
                        else
                            result = ITZAM_OKAY;
                    }
                }
            }
        }
    }
    
    // can size_t hold the types of references used by this file?
    if (result == ITZAM_OKAY)
    {
        size_t refbytes = itzam_datafile_get_refbits(datafile) / 8;
        
        if (sizeof(size_t) < refbytes)
            result = ITZAM_REF_SIZE_MISMATCH;
    }

    // done
    return result;
}

itzam_state itzam_datafile_close(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // close file
        if (0 == fclose(datafile->m_file))
        {
            datafile->m_file = NULL;
            result = ITZAM_OKAY;
        }
        else
            datafile->m_error_handler("itzam_datafile_close",ITZAM_ERROR_CLOSE_FAILED);
    }

    // done
    return result;
}

int itzam_datafile_get_refbits(itzam_datafile * datafile)
{
    return (int)((datafile->m_header.m_version & 0xFF000000) >> 24);
}

float itzam_datafile_get_version(itzam_datafile * datafile)
{
    return (double)((datafile->m_header.m_version & 0x00FF0000) >> 16)
            + ((datafile->m_header.m_version & 0x0000FF00) >> 8) / 100.0
            + ((datafile->m_header.m_version & 0x000000FF)) / 10000.0;
}

void itzam_datafile_set_error_handler(itzam_datafile * datafile,
                                      itzam_error_handler * error_handler)
{
    if (datafile != NULL)
    {
        if (error_handler == NULL)
            datafile->m_error_handler = itzam_get_default_error_handler();
        else
            datafile->m_error_handler = error_handler;
    }
}

itzam_ref itzam_datafile_tell(itzam_datafile * datafile)
{
    // what we return
    itzam_ref where = -1;

    // verify arguments before proceeding
    if (datafile != NULL)
        where = (itzam_ref)ftell(datafile->m_file);

    //done
    return where;
}

itzam_state itzam_datafile_seek(itzam_datafile * datafile, itzam_ref  pos)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // seek to requested position
        if (0 == fseek(datafile->m_file,(long)pos,SEEK_SET))
            result = ITZAM_OKAY;
        else
            datafile->m_error_handler("itzam_datafile_seek",ITZAM_ERROR_SEEK_FAILED);
    }

    // done
    return result;
}

itzam_state itzam_datafile_rewind(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // seek to byte after first deleted marker
        if (0 == fseek(datafile->m_file,(long)sizeof(itzam_datafile_header),SEEK_SET))
            result = ITZAM_OKAY;
        else
            datafile->m_error_handler("itzam_datafile_rewind",ITZAM_ERROR_SEEK_FAILED);
    }

    // done
    return result;
}

itzam_state itzam_datafile_skip(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // read length of current record
        itzam_record_header header;

        if (1 == fread(&header,sizeof(header),1,datafile->m_file))
        {
            // skip to beginning of next record
            if (0 == fseek(datafile->m_file,(long)header.m_length,SEEK_CUR))
                result = ITZAM_OKAY;
            else
                datafile->m_error_handler("itzam_datafile_skip",ITZAM_ERROR_SEEK_FAILED);
        }
        else
            datafile->m_error_handler("itzam_datafile_skip",ITZAM_ERROR_READ_FAILED);
    }

    // done
    return result;
}

// this function can be a tad dangerous, as it assumes that a returned deleted record will
//      be used in the calling function...
itzam_ref itzam_datafile_get_next_open(itzam_datafile * datafile, itzam_int length)
{
    // what we return
    itzam_int n;
    itzam_ref where = ITZAM_NULL_REF;

    // make sure the arguments make sense
    if ((datafile != NULL) && (length > 0))
    {
        if (datafile->m_header.m_dellist_ref != ITZAM_NULL_REF)
        {
            for (n = 0; n < datafile->m_dellist_header.m_table_size; ++n)
            {
                if (datafile->m_dellist[n].m_where != ITZAM_NULL_REF)
                {
                    if (datafile->m_dellist[n].m_length == length)
                    {
                        // save thge location of the deleted record
                        where = datafile->m_dellist[n].m_where;

                        // remove this entry from the table
                        datafile->m_dellist[n].m_where  = ITZAM_NULL_REF;
                        datafile->m_dellist[n].m_length = 0;
                        write_dellist(datafile,false);
                        break;
                    }
                }
            }
        }
        
        if (where == ITZAM_NULL_REF)
        {
            // no deleted records, so append
            if (0 == fseek(datafile->m_file,0,SEEK_END))
            {
                where = ftell(datafile->m_file);

                if (where < 0)
                    where = ITZAM_NULL_REF; // where is probably -1 if we get here, but just to be complete and symbolic...
            }
            else
                where = ITZAM_NULL_REF; // as above
        }
    }

    // done
    return where;
}

itzam_ref itzam_datafile_write(itzam_datafile * datafile, const void * data, itzam_int length, itzam_ref where)
{
    itzam_ref prev = ITZAM_NULL_REF;
    itzam_record_header header;
    
    // make sure the arguments make sense
    if ((datafile != NULL) && (data != NULL) && (length > 0))
    {
        if (where == ITZAM_NULL_REF)
            where = itzam_datafile_get_next_open(datafile,length);
        
        header.m_flags    = ITZAM_RECORD_IN_USE;
        header.m_length   = 0;
        header.m_rec_size = 0;
        
        // do we have a good place to write?
        if (where != ITZAM_NULL_REF)
        {
            // modify record header
            header.m_rec_size = length;

            if (header.m_length == 0)
                header.m_length = length;

            // seek to beginning of new record
            if (0 == fseek(datafile->m_file,where,SEEK_SET))
            {
                // write the record header
                if (1 == fwrite(&header,sizeof(header),1,datafile->m_file))
                {
                    // write the record
                    if (1 != fwrite(data,length,1,datafile->m_file))
                    {
                        datafile->m_error_handler("itzam_datafile_write",ITZAM_ERROR_WRITE_FAILED);
                        where = ITZAM_NULL_REF;
                    }
                }
                else
                {
                    datafile->m_error_handler("itzam_datafile_write",ITZAM_ERROR_WRITE_FAILED);
                    where = ITZAM_NULL_REF;
                }
            }
            else
            {
                datafile->m_error_handler("itzam_datafile_write",ITZAM_ERROR_SEEK_FAILED);
                where = ITZAM_NULL_REF;
            }
        }
    }

    // done
    return where;
}

itzam_state itzam_datafile_rewrite(itzam_datafile * datafile, const void * data, itzam_int length, itzam_ref where)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((datafile != NULL) && (data != NULL) && (length > 0))
    {
        if (0 == (itzam_ref)fseek(datafile->m_file,where,SEEK_SET))
        {
            itzam_record_header header;

            // read the header
            if (1 == fread(&header,sizeof(header),1,datafile->m_file))
            {
                // verify that there is enough room, and that we're replacing an existing record
                if (header.m_length <= length)
                {
                    if (header.m_flags & ITZAM_RECORD_IN_USE)
                    {
                        // update length
                        header.m_rec_size = length;

                        // find the header again
                        fseek(datafile->m_file,where,SEEK_SET); // this shouldn't fail, since it just worked!

                        // update header
                        if (1 == fwrite(&header,sizeof(header),1,datafile->m_file))
                        {
                            if (1 == fwrite(data,length,1,datafile->m_file))
                                result = ITZAM_OKAY;
                            else
                                datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_WRITE_FAILED);
                        }
                        else
                        {
                            datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_WRITE_FAILED);
                        }
                    }
                    else
                    {
                        datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_REWRITE_DELETED);
                        where = ITZAM_NULL_REF;
                    }
                }
                else
                {
                    datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_TOO_SMALL);
                    where = ITZAM_NULL_REF;
                }
            }
            else
            {
                datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_READ_FAILED);
                where = ITZAM_NULL_REF;
            }
        }
        else
        {
            datafile->m_error_handler("itzam_datafile_rewrite",ITZAM_ERROR_SEEK_FAILED);
            where = ITZAM_NULL_REF;
        }
    }

    // done
    return result;
}

itzam_state itzam_datafile_read(itzam_datafile * datafile, void * data, itzam_int max_length)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((datafile != NULL) && (data != NULL) && (max_length > 0))
    {
        // working storage
        itzam_record_header header;
        bool error = false;

        while (!error)
        {
            // read the record header
            if (1 == fread(&header,sizeof(header),1,datafile->m_file))
            {
                // if the record is active, we can read it
                if ((header.m_flags & ITZAM_RECORD_IN_USE) && !(header.m_flags & ITZAM_RECORD_DELLIST))
                    break;

                // move to the next record
                if (fseek(datafile->m_file,header.m_length,SEEK_CUR))
                {
                    datafile->m_error_handler("itzam_datafile_read",ITZAM_ERROR_SEEK_FAILED);
                    error = true;
                }
            }
            else
            {
                datafile->m_error_handler("itzam_datafile_read",ITZAM_ERROR_READ_FAILED);
                error = true;
            }
        }

        // only read if no errors
        if (!error)
        {
            // create a temporary buffer
            char * buffer = (char *)malloc(header.m_rec_size);

            // read data
            if (1 == fread(buffer,header.m_rec_size,1,datafile->m_file))
            {
                // skip any "padding" between record size and record buffer length
                if (header.m_rec_size < header.m_length)
                {
                    if (0 != fseek(datafile->m_file,header.m_length - header.m_rec_size,SEEK_CUR))
                    {
                        datafile->m_error_handler("itzam_datafile_read",ITZAM_ERROR_SEEK_FAILED);
                        error = true;
                    }
                }
            }
            else
            {
                datafile->m_error_handler("itzam_datafile_read",ITZAM_ERROR_READ_FAILED);
                error = true;
            }

            // watch for error in possible seek above
            if (!error)
            {
                // don't copy more data than caller says we can
                memcpy(data,buffer,(max_length < header.m_rec_size) ? max_length : header.m_rec_size);
                result = ITZAM_OKAY;
            }

            // release buffer
            free(buffer);
        }
    }

    // done
    return result;
}

itzam_state itzam_datafile_read_alloc(itzam_datafile * datafile, void ** data, itzam_ref * length)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;

    // make sure the arguments make sense
    if ((datafile != NULL) && (data != NULL))
    {
        // working storage
        itzam_record_header header;
        bool error = false;

        while (!error)
        {
            // read the record header
            if (1 == fread(&header,sizeof(header),1,datafile->m_file))
            {
                // if the record is active, we can read it
                if (header.m_flags & ITZAM_RECORD_IN_USE)
                    break;

                // move to the next record
                if (fseek(datafile->m_file,header.m_length,SEEK_CUR))
                {
                    datafile->m_error_handler("itzam_datafile_read_alloc",ITZAM_ERROR_SEEK_FAILED);
                    error = true;
                }
            }
            else
            {
                datafile->m_error_handler("itzam_datafile_read_alloc",ITZAM_ERROR_READ_FAILED);
                error = true;
            }
        }

        // only read if no errors
        if (!error)
        {
            // create a temporary buffer
            *length = header.m_rec_size;
            *data = malloc(header.m_rec_size);

            // read data
            if (*data != NULL)
            {
                if (1 == fread(*data,header.m_rec_size,1,datafile->m_file))
                {
                    // skip any "padding" between record size and record buffer length
                    if (header.m_rec_size < header.m_length)
                        fseek(datafile->m_file,header.m_length - header.m_rec_size,SEEK_CUR);

                    result = ITZAM_OKAY;
                }
                else
                    datafile->m_error_handler("itzam_datafile_read_alloc",ITZAM_ERROR_READ_FAILED);
            }
            else
                datafile->m_error_handler("itzam_datafile_read_alloc",ITZAM_ERROR_NULL_ALLOC);
        }
    }

    // done
    return result;
}

itzam_state itzam_datafile_remove(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_UNKNOWN;
    int n;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        // get our position
        itzam_ref where = ftell(datafile->m_file);
        
        // make sure we're not in the file header
        if (where >= sizeof(itzam_ref))
        {
            itzam_record_header header;

            // read the header
            if (1 == fread(&header,sizeof(header),1,datafile->m_file))
            {
                // only delete if it isn't already deleted
                if (header.m_flags & ITZAM_RECORD_IN_USE)
                {
                    // change record header; make this record the head of the deleted list
                    header.m_flags &= ~ITZAM_RECORD_IN_USE;

                    // move to beginning of record header
                    if (0 == fseek(datafile->m_file,where,SEEK_SET))
                    {
                        // write revised record header
                        if (1 == fwrite(&header,sizeof(header),1,datafile->m_file))
                        {
                            // update deleted list
                            if (datafile->m_header.m_dellist_ref == ITZAM_NULL_REF)
                            {
                                datafile->m_dellist_header.m_table_size = ITZAM_DEL_LIST_BLOCK_SIZE;
                                datafile->m_dellist = (itzam_dellist_entry *)malloc(sizeof(itzam_dellist_entry) * datafile->m_dellist_header.m_table_size);
                                
                                if (datafile->m_dellist != NULL)
                                {
                                    datafile->m_dellist[0].m_where  = where;
                                    datafile->m_dellist[0].m_length = header.m_length;
                                    
                                    for (n = 1; n < datafile->m_dellist_header.m_table_size; ++n)
                                    {
                                        datafile->m_dellist[n].m_where  = ITZAM_NULL_REF;
                                        datafile->m_dellist[n].m_length = 0;
                                    }
                                    
                                    result = write_dellist(datafile,true);
                                }
                            }
                            else
                            {
                                for (n = 0; n < datafile->m_dellist_header.m_table_size; ++n)
                                {
                                    if (datafile->m_dellist[n].m_where == ITZAM_NULL_REF)
                                    {
                                        datafile->m_dellist[n].m_where  = where;
                                        datafile->m_dellist[n].m_length = header.m_length;
                                        result = write_dellist(datafile,false);
                                        break;
                                    }
                                }

                                // if the entry didn't fit, we have to expand it
                                if (n == datafile->m_dellist_header.m_table_size)
                                {
                                    // create new table
                                    itzam_int newsize = datafile->m_dellist_header.m_table_size + ITZAM_DEL_LIST_BLOCK_SIZE;
                                    itzam_dellist_entry * newlist = (itzam_dellist_entry *)malloc(sizeof(itzam_dellist_entry) * newsize);
                                    
                                    // copy old entries; could (?should) use memcpy here I suppose
                                    for (n = 0; n < datafile->m_dellist_header.m_table_size; ++n)
                                    {
                                        newlist[n].m_where  = ITZAM_NULL_REF;
                                        newlist[n].m_length = 0;
                                    }
                                    
                                    // add the newly-deleted record
                                    newlist[n].m_where  = where;
                                    newlist[n].m_length = header.m_length;
                                    
                                    // add the old deleted list record
                                    newlist[n+1].m_where  = datafile->m_header.m_dellist_ref;
                                    newlist[n+1].m_length = datafile->m_dellist_header.m_table_size;
                                    
                                    // initialize remaining unused new entries
                                    for (n += 2 ; n < newsize; ++n)
                                    {
                                        newlist[n].m_where  = datafile->m_dellist[n].m_where;
                                        newlist[n].m_length = datafile->m_dellist[n].m_length;
                                    }
                                    
                                    // free space used by old list
                                    free(datafile->m_dellist);
                                    
                                    // exchange new list for old
                                    datafile->m_dellist = newlist;
                                    datafile->m_dellist_header.m_table_size = newsize;
                                    
                                    // write it
                                    result = write_dellist(datafile,true);
                                }
                            }
                        }
                        else
                            datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_WRITE_FAILED);
                    }
                    else
                        datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_SEEK_FAILED);
                }
                else
                    datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_DUPE_REMOVE);
            }
            else
                datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_READ_FAILED);
        }
        else
            datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_TELL_FAILED);
    }

    // done
    return result;
}

// flush (commit) data to disk
itzam_state itzam_datafile_commit(itzam_datafile * datafile)
{
    // what we return
    itzam_state result = ITZAM_OKAY;

    // verify arguments before proceeding
    if (datafile != NULL)
    {
        if (0 != fflush(datafile->m_file))
        {
            datafile->m_error_handler("itzam_datafile_remove",ITZAM_ERROR_FLUSH_FAILED);
            result = ITZAM_UNKNOWN;
        }
    }

    // done
    return result;
}

