//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam_bigfile.c
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

//----------------------------------------------------------
// Standard C
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

#if !defined(_MSC_VER)
#include <fcntl.h>
#include <unistd.h>
#endif

//----------------------------------------------------------
// itzam
#include "../libitzam/itzam.h"
#include "../libitzam/itzam_debug.h"

//----------------------------------------------------------
//  An index uses functions of this form to compare keys.
//  Functions of this type must return:
//      < 0    key1 is before key2
//        0    keys are equal
//      > 0    key1 is after key2

static int comparator(const void * key1, const void * key2)
{
    return (*((int64_t *)key1)) - (*((int64_t *)key2));
}

//----------------------------------------------------------
//  Function to select keys to be included in an iterator.
//  The meaning of the return value is:
//      true    include key
//      false   don't include this key
static bool selector(const void * key)
{
    // select all keys
    return true;
}

//----------------------------------------------------------
//  Reports an itzam error
static void not_okay(itzam_state state)
{
    printf("\nProblematic Itzam state: %s\n",itzam_get_state_description(state));
    exit(EXIT_FAILURE);
}

//----------------------------------------------------------
//  Display program usage
static const char * USAGE =
    "\nusage: itzam_bigfile [options]\n\n"
    "options:\n"
    "    -recsize n size of a record, in kilobytes (default = 1,024)\n"
    "    -numrecs n number of records to write (default = 4100)\n"
    "    -help      this usage information\n"
    "\n";

static void show_usage()
{
    puts(USAGE);
}

//----------------------------------------------------------
// main program
int main(int argc, char * argv[])
{
    // test code
    itzam_btree_iterator iterator;
    itzam_btree btree;
    itzam_state state;
    itzam_ref   ref;
    itzam_int   key, rec_key;
    itzam_int   numrecs  = 4100;
    itzam_int   recsize  = 1024 * 1024;
    int         i;
    uint8_t *   record;

    // generate a random file name
    uint32_t file_code = 0;
    const size_t MAX_FILE_NAME = 64;
    char file_name[64]; // MS can't handle using constant above

#if defined(_MSC_VER) && (_MSC_VER < 1400)
    sprintf(file_name,"bigfile%08x.dat",time(NULL));
#elif defined(_MSC_VER) && (_MSC_VER >= 1400)
    sprintf_s(file_name,MAX_FILE_NAME,"bigfile%08x.dat",time(NULL));
#else
    // first try to read /dev/urandom
    int fd = open ("/dev/urandom", O_RDONLY);

    if (fd != -1)
    {
        read(fd, &file_code, 4);
        close(fd);
    }
    else
        file_code = (uint32_t)time(NULL);

    snprintf(file_name,MAX_FILE_NAME,"bigfile%08x.dat",file_code);
#endif

    // adjust number of records based on int size
    switch (sizeof(int))
    {
        case 4:
            numrecs = 2047; // just under 2GB for 32-bit signed file pointers
            break;
        case 8:
            numrecs = 5000; // over 4GB for 64-bit signed file pointers
            break;
        default:
            numrecs = 1000;
            break;
    }

    // parse arguments
    for (i = 1; i < argc; ++i)
    {
        if (0 == strcmp(argv[i],"-help"))
        {
            show_usage();
            exit(EXIT_FAILURE);
        }
        else if (0 == strcmp(argv[i],"-numrecs"))
        {
            ++i;

            if (i < argc)
                numrecs = strtoul(argv[i],NULL,10);

            if (numrecs < 1)
                numrecs = 1;
        }
        else if (0 == strcmp(argv[i],"-recsize"))
        {
            ++i;

            if (i < argc)
                recsize = strtoul(argv[i],NULL,10) * 1024;

            if (recsize < 1024)
                recsize = 1024;
        }
    }

    // create database
    state = itzam_btree_create(&btree,file_name,comparator);

    if (state != ITZAM_OKAY)
        not_okay(state);

    printf("done\n");
    fflush(stdout);

    // allocate a record
    record = (uint8_t *)malloc((size_t)recsize);

    if (record == NULL)
    {
        printf("memory allocation failed for record of size %ld\n",recsize);
        exit(EXIT_FAILURE);
    }

    // clear record
    memset((void *)record,0,(size_t)recsize);

    // write lots of records
    for (key = 1; key <= numrecs; ++ key)
    {
        // copy record number into initial bytes for later verification
        memcpy((void *)record,(const void *)&key,sizeof(key));

        // display key and loop number
        printf("writing %10ld: ",key);

        // write record
        ref = itzam_datafile_write(btree.m_datafile,(const void *)record,recsize,ITZAM_NULL_REF);

        if (ref == ITZAM_NULL_REF)
        {
            printf("error, could not write record\n");
            exit(EXIT_FAILURE);
        }

        // insert record referemce into tree
        state = itzam_btree_insert(&btree,(void *)(&key),sizeof(key),ref);

        if (state == ITZAM_OKAY)
        {
            printf("written\n");
        }
        else
        {
            printf("error, could not index record\n");
            exit(EXIT_FAILURE);
        }
    }

    // now iterate the database
    itzam_btree_iterator_create(&btree,&iterator,NULL);

    if (itzam_btree_iterator_is_valid(&iterator))
    {
        key = 1;

        do
        {
            // get the record reference
            printf("reading %10ld: ",key);
            ref = itzam_btree_iterator_get_rec_ref(&iterator);

            // find the record
            state = itzam_datafile_seek(btree.m_datafile,ref);

            if (state != ITZAM_OKAY)
            {
                printf("error, could seek record\n");
                exit(EXIT_FAILURE);
            }

            // read the record
            state = itzam_datafile_read(btree.m_datafile,(void *)record,recsize);

            if (state != ITZAM_OKAY)
            {
                printf("error, could read record\n");
                exit(EXIT_FAILURE);
            }

            // extract the key to see if it matches
            memcpy((void *)&rec_key,(const void *)record,sizeof(rec_key));

            printf("found %10ld, ", rec_key);

            // check to see if read data matches key
            if (rec_key == key)
                printf("good\n");
            else
            {
                printf("bad!\n");
                exit(EXIT_FAILURE);
            }

            ++key;
        }
        while (itzam_btree_iterator_move_next(&iterator) == ITZAM_OKAY);
    }

    itzam_btree_iterator_free(&iterator);

    printf("\n");

    // close database
    itzam_btree_close(&btree);

    // done
    return EXIT_SUCCESS;
}
