//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  Itzam - An Embedded Database Engine
//
//  itzam_exercise.c
//
//---------------------------------------------------------------------
//
//  Copyright 2004, 2005 Scott Robert Ladd
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the
//      Free Software Foundation, Inc.
//      59 Temple Place - Suite 330
//      Boston, MA 02111-1307, USA.
//
//-----------------------------------------------------------------------
//
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//
//-----------------------------------------------------------------------

//----------------------------------------------------------
// Standard C
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

#if !defined(_MSC_VER)
#include <fcntl.h>
#include <unistd.h>
#endif

//----------------------------------------------------------
// itzam
#include "../libitzam/itzam.h"
#include "../libitzam/itzam_debug.h"

//----------------------------------------------------------
// embedded random number generator; ala Park and Miller
static int32_t seed = 1325;

static int32_t random_int32(int32_t limit)
{
    static const int32_t IA   = 16807;
    static const int32_t IM   = 2147483647;
    static const int32_t IQ   = 127773;
    static const int32_t IR   = 2836;
    static const int32_t MASK = 123459876;

    int32_t k;
    int32_t result;

    seed ^= MASK;
    k = seed / IQ;
    seed = IA * (seed - k * IQ) - IR * k;

    if (seed < 0L)
        seed += IM;

    result = (seed % limit);
    seed ^= MASK;

    return result;
}

//----------------------------------------------------------
//  An index uses functions of this form to compare keys.
//  Functions of this type must return:
//      < 0    key1 is before key2
//        0    keys are equal
//      > 0    key1 is after key2

static int comparator(const void * key1, const void * key2)
{
    return (*((int64_t *)key1)) - (*((int64_t *)key2));
}

//----------------------------------------------------------
//  Function to select keys to be included in an iterator.
//  The meaning of the return value is:
//      true    include key
//      false   don't include this key
static bool selector(const void * key)
{
    // select odd keys
    return (((*((int64_t *)key)) & 1) == 1);
}

//----------------------------------------------------------
//  Verifies that the database contains the expected records
static void verify(itzam_btree * btree, bool * key_flags, int max_key)
{
    itzam_state state;
    int64_t key, rec, ref;

    // make sure check values match btree
    printf(" -- verifying... ");

    for (key = 0; key < max_key; ++key)
    {
        ref = itzam_btree_find(btree,(const void *)(&key));

        if (ref == ITZAM_NULL_REF)
        {
            if (key_flags[key])
            {
                printf("expected key not found\n");
                exit(EXIT_FAILURE);
            }
        }
        else
        {
            if (key_flags[key])
            {
                state = itzam_btree_read_rec(btree,(const void *)(&key),(void *)(&rec),sizeof(rec));

                if (state == ITZAM_OKAY)
                {
                    if (rec != key)
                    {
                        printf("data does not match key\n");
                        exit(EXIT_FAILURE);
                    }
                }
                else
                {
                    printf("unable to read existing record\n");
                    exit(EXIT_FAILURE);
                }
            }
            else
            {
                printf("key found, and should not have been\n");
                exit(EXIT_FAILURE);
            }
        }
    }

    printf("okay\n");
}

//----------------------------------------------------------
//  Reports an itzam error
static void not_okay(itzam_state state)
{
    printf("\nProblematic Itzam state: %s\n",itzam_get_state_description(state));
    exit(EXIT_FAILURE);
}

//----------------------------------------------------------
//  Display program usage
static const char * USAGE =
    "\nusage: itzam_exercise [options]\n\n"
    "options:\n"
    "    -tests n   number of test iterations to run (default = 10000)\n"
    "    -maxkey n  maximum key value (default = 10000)\n"
    "    -verify    verify database contents after every operation (slow)\n"
    "    -seed n    use n as the random number seed (default is random)\n"
    "    -list      at the end, list the database using iterators\n"
    "    -dump      dump the database structure to stdout (very verbose)\n"
    "    -analyze   analyze the structure of the tree, giving statistics\n"
    "    -verbose   display progress text\n"
    "    -help      this usage information\n"
    "\n";

static void show_usage()
{
    puts(USAGE);
}

//----------------------------------------------------------
// main program
int main(int argc, char * argv[])
{
    // test code
    size_t      n;
    itzam_btree btree;
    itzam_state state;
    itzam_ref   ref;
    int64_t     key;
    int         i;
    bool *      key_flags;
    itzam_int   add_count = 0;
    itzam_int   rem_count = 0;

    uint64_t max_key  = 10000;
    size_t test_size  = 10000;
    bool verified_run = false;
    bool iter_list    = false;
    bool dump_db      = false;
    bool analyze      = false;
    bool verbose      = false;
    const char * prog = "|/-\\";

    // generate a random file name
    uint32_t file_code = 0;
    const size_t MAX_FILE_NAME = 64;
    char file_name[64]; // MS can't handle using constant above

#if defined(_MSC_VER) && (_MSC_VER < 1400)
    sprintf(file_name,"btree%08x.dat",time(NULL));
#elif defined(_MSC_VER) && (_MSC_VER >= 1400)
    sprintf_s(file_name,MAX_FILE_NAME,"btree%08x.dat",time(NULL));
#else
    // first try to read /dev/urandom
    int fd = open ("/dev/urandom", O_RDONLY);

    if (fd != -1)
    {
        read(fd, &file_code, 4);
        close(fd);
    }
    else
        file_code = (uint32_t)time(NULL);

    snprintf(file_name,MAX_FILE_NAME,"btree%08x.dat",file_code);
#endif

    // parse command-line arguments
    seed = (long)time(NULL);

    // parse arguments
    for (i = 1; i < argc; ++i)
    {
        if (0 == strcmp(argv[i],"-verify"))
        {
            verified_run = true;
        }
        else if (0 == strcmp(argv[i],"-help"))
        {
            show_usage();
            exit(EXIT_FAILURE);
        }
        else if (0 == strcmp(argv[i],"-list"))
        {
            iter_list = true;
        }
        else if (0 == strcmp(argv[i],"-dump"))
        {
            dump_db = true;
        }
        else if (0 == strcmp(argv[i],"-analyze"))
        {
            analyze = true;
        }
        else if (0 == strcmp(argv[i],"-verbose"))
        {
            verbose = true;
        }
        else if (0 == strcmp(argv[i],"-seed"))
        {
            ++i;

            if (i < argc)
                seed = strtol(argv[i],NULL,10);
        }
        else if (0 == strcmp(argv[i],"-tests"))
        {
            ++i;

            if (i < argc)
                test_size = strtoul(argv[i],NULL,10);
        }
        else if (0 == strcmp(argv[i],"-maxkey"))
        {
            ++i;

            if (i < argc)
                max_key = strtoul(argv[i],NULL,10);
        }
        else
        {
            printf("unknown option: %s\n",argv[i]);
            show_usage();
            exit(EXIT_FAILURE);
        }
    }

    // create database
    state = itzam_btree_create(&btree,file_name,comparator);

    if (state != ITZAM_OKAY)
        not_okay(state);

    // describe parameters of the run
    printf("itzam version....... %7.4f\nreference bits...... %d\n",
            itzam_datafile_get_version(btree.m_datafile),
            itzam_datafile_get_refbits(btree.m_datafile));
    
    printf("number of tests..... %lu\nmaximum key value... %lu\nrandom seed......... %ld\n\n",
            test_size, max_key, seed);

    // allocate array of flag to track which keys are supposed to be in the database
    key_flags = (bool *)malloc(sizeof(bool) * max_key);

    for (n = 0; n < max_key; ++n)
        key_flags[n] = false;

    // stress test
    for (n = 1; n <= test_size; ++ n)
    {
        // pick random key
        key = (int64_t)random_int32((size_t)max_key);

        // display key and loop number
        if (verbose)
            printf("%8lu: test key = %8lu: ",n,key);
        else
        {
            if (0 == n % 1000)
                printf("\b. ");
            else
                printf("\b%c",prog[n % 4]);

            fflush(stdout);
        }
            
        // see if this key already exists
        state = itzam_btree_insert_rec(&btree,
                                       (void *)&key,sizeof(key),
                                       (void *)&key,sizeof(key));

        switch (state)
        {
            case ITZAM_OKAY:
                if (key_flags[key])
                {
                    printf("key was not found, and should have been\n");
                    exit(EXIT_FAILURE);
                }
                
                key_flags[key] = true;
                ++add_count;

                if (verbose)
                    printf("written successfully");
                
                break;
                
            case ITZAM_DUPLICATE:
                if (!key_flags[key])
                {
                    printf("key was found, and should not have been\n");
                    exit(EXIT_FAILURE);
                }
                
                state = itzam_btree_remove(&btree,(void *)(&key),NULL);
                ++rem_count;
                
                if (state == 0)
                {
                    key_flags[key] = false;
                    
                    if (verbose)
                        printf("removed successfully");
                }
                else
                {
                    printf("unable to remove data\n");
                    exit(EXIT_FAILURE);
                }

                break;
                
            default:
                not_okay(state);
        }

        // make certain nothing has broken
        if (verified_run)
            verify(&btree,key_flags,max_key);
        else
           if (verbose)
                printf(" -- done\n");

        // flush output line
        fflush(stdout);
    }
    
    // display number of record added and removed
    printf("\n\ntotal records   added: %ld\n",add_count);
    printf("total records removed: %ld\n",rem_count);

    // verify database after all operations
    printf("\ndatabase check ");
    verify(&btree,key_flags,max_key);

    // now iterate the database
    if (iter_list)
    {
        // all keys forward
        itzam_btree_iterator iterator;

        itzam_btree_iterator_create(&btree,&iterator,NULL);

        printf("\n\nIterating all keys forward:\n");

        if (itzam_btree_iterator_is_valid(&iterator))
        {
            do
            {
                printf("%d ",(int32_t)itzam_btree_iterator_get_rec_ref(&iterator));
                state = itzam_btree_iterator_move_next(&iterator);
            }
            while (state == ITZAM_OKAY);
        }

        // now all keys backward
        printf("\n\nIterating all keys backward:\n");

        itzam_btree_iterator_move_last(&iterator);

        if (itzam_btree_iterator_is_valid(&iterator))
        {
            do
            {
                printf("%d ",(int32_t)itzam_btree_iterator_get_rec_ref(&iterator));
                state = itzam_btree_iterator_move_prev(&iterator);
            }
            while (state == ITZAM_OKAY);
        }

        itzam_btree_iterator_free(&iterator);

        // all odd keys
        itzam_btree_iterator_create(&btree,&iterator,selector);

        printf("\n\nIterating all odd keys forward:\n");

        if (itzam_btree_iterator_is_valid(&iterator))
        {
            do
            {
                printf("%d ",(int32_t)itzam_btree_iterator_get_rec_ref(&iterator));
                state = itzam_btree_iterator_move_next(&iterator);
            }
            while (state == ITZAM_OKAY);
        }

        itzam_btree_iterator_free(&iterator);
    }

    printf("\n");

    // dump the btree structure, if requested
    if (dump_db)
        dump_btree(&btree);

    // analyze the btree structure, if requested
    if (analyze)
        analyze_btree(&btree,stdout);

    // clean up
    free(key_flags);

    // close database
    itzam_btree_close(&btree);

    // done
    return EXIT_SUCCESS;
}
